# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireConversationResolution object."""

import textwrap

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RequireConversationResolution(ClassicEnableRequirementImpl):
    """Require conversation resolution before merging."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequireConversationResolution",
            True,
            "disabled",
            "Protect matching branches",
            "Require conversation resolution before merging",
            lambda data: data["branch_protection_data"]["required_conversation_resolution"]["enabled"],
            textwrap.dedent(
                """\
                The default behavior is to require conversation resolution before merging a pull request.

                Reasons for this Default
                ------------------------
                - Conversation resolution is an important part of the development process.
                - Prevent the accidental merging of a pull request before changes associated with the comments have been made.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireSignedCommits object."""

import textwrap

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RequireSignedCommits(ClassicEnableRequirementImpl):
    """Require signed commits."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequireSignedCommits",
            True,
            "disabled",
            "Protect matching branches",
            "Require signed commits",
            lambda data: data["branch_protection_data"]["required_signatures"]["enabled"],
            textwrap.dedent(
                """\
                The default behavior is to require signed commits. Note that this setting does not work with
                rebase merging or squash merging.

                Reasons for this Default
                ------------------------
                - Ensure that the author of a commit is who the claim to be.

                Reasons to Override this Default
                --------------------------------
                - You have enabled rebase merging or squash merging.
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the Private object."""

import textwrap
from typing import Any

import typer
from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.GitHub.Impl.Common import CreateIncompleteDataResult
from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement


# ----------------------------------------------------------------------
class Private(Requirement):
    """Validates that the repository is set to the expected visibility."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "Private",
            "Validates that the repository is set to the expected visibility.",
            ExecutionStyle.Parallel,
            textwrap.dedent(
                """\
                1) Visit '{session.github_url}/settings'
                2) Locate the 'Danger Zone' section
                3) Click the 'Change visibility button'
                4) Select 'Change to {__expected_visibility}'
                """,
            ),
            textwrap.dedent(
                """\
                The default behavior is to ensure that the repository is public.

                Reasons for this Default
                ------------------------
                - The vast majority of repositories hosted by GitHub are public.

                Reasons to Override this Default
                --------------------------------
                - Your repository really should be private.
                """,
            ),
        )

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Get the definitions for the arguments to this requirement."""
        return {
            "enabled": (
                bool,
                typer.Option(
                    default=False,
                    help="Ensures that the repository's visibility is set to private.",
                ),
            ),
        }

    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        result = query_data["standard"].get("private", None)

        if result is None:
            return CreateIncompleteDataResult()

        expect_private = requirement_args["enabled"]

        if result != expect_private:
            query_data["__expected_visibility"] = "private" if expect_private else "public"

            return Requirement.EvaluateImplResult(
                EvaluateResult.Error,
                f"The repository's visibility must be {query_data['__expected_visibility']}.",
                provide_resolution=True,
                provide_rationale=not expect_private,
            )

        return Requirement.EvaluateImplResult(EvaluateResult.Success, None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RulesetQuery object."""

from typing import Any, Optional

from dbrownell_Common.Types import override

from RepoAuditor.Impl.ParallelSequentialProcessor import ExecutionStyle
from RepoAuditor.Plugins.GitHub.RulesetRequirements.RequirePullRequests import RequirePullRequests
from RepoAuditor.Plugins.GitHub.RulesetRequirements.RequireSignedCommits import RequireSignedCommits
from RepoAuditor.Plugins.GitHub.RulesetRequirements.RequireStatusChecks import RequireStatusChecks
from RepoAuditor.Query import Query


class RulesetQuery(Query):
    """Query to validate GitHub repository rulesets."""

    def __init__(self) -> None:
        super().__init__(
            "RulesetQuery",
            ExecutionStyle.Parallel,
            [
                RequireSignedCommits(),
                RequirePullRequests(),
                RequireStatusChecks(),
            ],
        )

    @override
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        """Retrieve data required for validating GitHub rulesets.

        Args:
            module_data (dict): Dictionary containing initial data from the module.

        Returns:
            Optional[dict]: A dictionary containing ruleset-related
            data, or None if an error occurs.

        """
        branch = module_data.get("branch")
        if branch is None:
            # Get the default branch name
            response = module_data["session"].get("")

            response.raise_for_status()
            response = response.json()

            branch = response["default_branch"]

        # Record the branch name
        module_data["branch"] = branch

        # Fetch ruleset data for a specific branch.
        # This only returns active rules on the branch
        rules_response = module_data["session"].get(f"rules/branches/{branch}")

        rules_response.raise_for_status()

        # Add ruleset data to module_data
        module_data["rules"] = rules_response.json()

        # Also get the associated ruleset for each rule
        for rule in module_data["rules"]:
            ruleset_respone = module_data["session"].get(f"rulesets/{rule['ruleset_id']}")
            ruleset_respone.raise_for_status()
            ruleset = ruleset_respone.json()
            rule["ruleset"] = ruleset

        return module_data



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit test for CommandLineProcessor.py"""

import re
import textwrap
from typing import Optional, cast
from unittest.mock import patch

import pytest
from dbrownell_Common.TestHelpers.StreamTestHelpers import GenerateDoneManagerAndContent
from dbrownell_Common.Types import override

from RepoAuditor.CommandLineProcessor import *
from RepoAuditor.Module import TypeDefinitionItemType
from RepoAuditor.Query import Query
from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
class MyModule(Module):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str = "MyModule",
        dynamic_args: Optional[dict[str, Any]] = None,
        *,
        requires_explicit_include: bool = False,
    ):
        super().__init__(
            name,
            "",
            ExecutionStyle.Sequential,
            [MyQuery()],
            requires_explicit_include=requires_explicit_include,
        )

        self.dynamic_args = dynamic_args

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        return self.dynamic_args or {}

    # ----------------------------------------------------------------------
    @override
    def GenerateInitialData(
        self,
        dynamic_args: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        assert not dynamic_args
        return dynamic_args


# ----------------------------------------------------------------------
class MyQuery(Query):
    # ----------------------------------------------------------------------
    def __init__(self):
        super().__init__(
            "MyQuery",
            ExecutionStyle.Sequential,
            [
                MyRequirement("Requirement1", requires_explicit_include=True, has_dynamic_args=False),
                MyRequirement("Requirement2", has_dynamic_args=True),
            ],
        )

    # ----------------------------------------------------------------------
    @override
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        return module_data


# ----------------------------------------------------------------------
class MyRequirement(Requirement):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        *,
        requires_explicit_include: bool = False,
        has_dynamic_args: bool = False,
    ):
        super().__init__(
            name,
            "",
            ExecutionStyle.Sequential,
            "",
            "",
            requires_explicit_include=requires_explicit_include,
        )

        self._has_dynamic_args = has_dynamic_args

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        if not self._has_dynamic_args:
            return {}

        return {"foo": int}

    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
    ) -> tuple[EvaluateResult, Optional[str]]:
        raise Exception("This should never be invoked")


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def test_Standard():
    clp = CommandLineProcessor.Create(
        lambda *args: {"MyModule-arg1": False},
        [MyModule()],
        [],
        [],
        set(),
        set(),
    )

    assert len(clp.module_infos) == 1
    assert isinstance(clp.module_infos[0].module, MyModule)
    assert clp.module_infos[0].module.GetNumRequirements() == 1
    assert clp.module_infos[0].dynamic_args == {"arg1": False}
    assert not clp.module_infos[0].requirement_args
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False

    dm_and_content = GenerateDoneManagerAndContent()

    with patch("RepoAuditor.CommandLineProcessor.Execute") as mock_execute:
        clp(next(dm_and_content))

    assert len(mock_execute.call_args_list) == 1

    args = mock_execute.call_args_list[0].args
    kwargs = mock_execute.call_args_list[0].kwargs

    assert len(args) == 2

    assert isinstance(args[0], DoneManager)
    assert args[1] == clp.module_infos

    assert kwargs == {}

    assert cast(str, next(dm_and_content)) == textwrap.dedent(
        """\
        Heading...DONE! (0, <scrubbed duration>)
        """,
    )


# ----------------------------------------------------------------------
def test_WithDynamicArgs():
    clp = CommandLineProcessor.Create(
        lambda *args: {
            "MyModule-arg1": False,
            "MyModule-one": 1,
            "MyModule-two": 2,
            "MyModule-Requirement2-foo": 123,
        },
        [MyModule(dynamic_args={"one": int, "two": int})],
        [],
        [],
        set(),
        set(),
    )

    assert len(clp.module_infos) == 1
    assert isinstance(clp.module_infos[0].module, MyModule)
    assert clp.module_infos[0].module.GetNumRequirements() == 1
    assert clp.module_infos[0].dynamic_args == {
        "arg1": False,
        "one": 1,
        "two": 2,
    }
    assert clp.module_infos[0].requirement_args == {
        "Requirement2": {"foo": 123},
    }
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
def test_IncludeModule():
    clp = CommandLineProcessor.Create(
        lambda *args: {},
        [MyModule(requires_explicit_include=True)],
        [],
        [],
        set(),
        set(),
    )

    assert not clp.module_infos
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False

    clp = CommandLineProcessor.Create(
        lambda *args: {},
        [MyModule(requires_explicit_include=True)],
        ["MyModule"],
        [],
        set(),
        set(),
    )

    assert len(clp.module_infos) == 1
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
def test_ExcludeModule():
    clp = CommandLineProcessor.Create(
        lambda *args: {},
        [MyModule()],
        [],
        ["MyModule"],
        set(),
        set(),
    )

    assert not clp.module_infos
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
@pytest.mark.parametrize("sep", ["-", "__--__"])
def test_IncludeRequirement(sep):
    clp = CommandLineProcessor.Create(
        lambda *args: {},
        [MyModule()],
        [f"MyModule{sep}Requirement1"],
        [],
        set(),
        set(),
        argument_separator=sep,
    )

    assert len(clp.module_infos) == 1
    assert clp.module_infos[0].module.GetNumRequirements() == 2
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
@pytest.mark.parametrize("sep", ["-", "__--__"])
def test_ExcludeRequirement(sep):
    clp = CommandLineProcessor.Create(
        lambda *args: {},
        [MyModule()],
        [],
        [f"MyModule{sep}Requirement2"],
        set(),
        set(),
        argument_separator=sep,
    )

    assert not clp.module_infos
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
def test_ExcludeAllRequirements():
    clp = CommandLineProcessor.Create(
        lambda *args: {},
        [MyModule()],
        [],
        ["MyModule-Requirement1", "MyModule-Requirement2"],
        set(),
        set(),
    )

    assert not clp.module_infos
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
@pytest.mark.parametrize("sep", ["-", "__--__"])
def test_CommandLineArgs(sep):
    clp = CommandLineProcessor.Create(
        lambda *args: {
            f"MyModule{sep}arg1": True,
        },
        [MyModule()],
        [],
        [],
        set(),
        set(),
        argument_separator=sep,
    )

    assert len(clp.module_infos) == 1
    assert clp.module_infos[0].module.GetNumRequirements() == 1
    assert clp.module_infos[0].dynamic_args == {"arg1": True}
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
def test_AllWarningsAsError():
    clp = CommandLineProcessor.Create(
        lambda *args: {"MyModule-arg1": False},
        [MyModule()],
        [],
        [],
        set(),
        set(),
        all_warnings_as_error=True,
    )

    assert len(clp.module_infos) == 1
    assert clp.warnings_as_error_module_names == {"MyModule"}
    assert clp.ignore_warnings_module_names == set()
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
def test_IgnoreAllWarnings():
    clp = CommandLineProcessor.Create(
        lambda *args: {"MyModule-arg1": False},
        [MyModule()],
        [],
        [],
        set(),
        set(),
        ignore_all_warnings=True,
    )

    assert len(clp.module_infos) == 1
    assert clp.warnings_as_error_module_names == set()
    assert clp.ignore_warnings_module_names == {"MyModule"}
    assert clp.single_threaded is False


# ----------------------------------------------------------------------
def test_ErrorInvalidModuleName():
    with pytest.raises(
        Exception,
        match=re.escape(
            "The module name 'Invalid-name' contains '-', which should be used as an argument separator."
        ),
    ):
        CommandLineProcessor.Create(
            lambda *args: {},
            [MyModule("Invalid-name")],
            [],
            [],
            set(),
            set(),
        )


# ----------------------------------------------------------------------
def test_ErrorDuplicateName():
    with pytest.raises(
        Exception,
        match=re.escape("The module 'MyModule' has already been defined."),
    ):
        CommandLineProcessor.Create(
            lambda *args: {},
            [MyModule(), MyModule()],
            [],
            [],
            set(),
            set(),
        )


# ----------------------------------------------------------------------
def test_ErrorInvalidIncludeName():
    with pytest.raises(
        Exception,
        match=re.escape("'DoesNotExist' is not a recognized module name."),
    ):
        CommandLineProcessor.Create(
            lambda *args: {"MyModule-arg1": False},
            [MyModule()],
            ["DoesNotExist"],
            [],
            set(),
            set(),
        )


# ----------------------------------------------------------------------
def test_ErrorInvalidExcludeName():
    with pytest.raises(
        Exception,
        match=re.escape("'DoesNotExist' is not a recognized module name."),
    ):
        CommandLineProcessor.Create(
            lambda *args: {"MyModule-arg1": False},
            [MyModule()],
            [],
            ["DoesNotExist"],
            set(),
            set(),
        )


# ----------------------------------------------------------------------
def test_InvalidArgName():
    with pytest.raises(
        Exception,
        match=re.escape("'MyModule' is not a recognized module name."),
    ):
        # This code provides an argument for a module that has been excluded
        CommandLineProcessor.Create(
            lambda *args: {"MyModule-arg1": False},
            [MyModule()],
            [],
            ["MyModule"],
            set(),
            set(),
        )


# ----------------------------------------------------------------------
def test_InvalidArgs():
    """Check if an invalid module is specified as an argument."""
    with pytest.raises(
        Exception,
        match=re.escape("'InvalidModule' is not a recognized module name."),
    ):
        CommandLineProcessor.Create(
            # Specify an unregistered module
            get_dynamic_args_func=lambda dynamic_arg_definitions: {"InvalidModule-arg1": "value"},
            # ValidModule is the registered module.
            modules=[MyModule(name="ValidModule")],
            includes=["InvalidModule"],
            excludes=[],
            warnings_as_error_module_names=set(),
            ignore_warnings_module_names=set(),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for Private.py"""

import pytest

from RepoAuditor.Plugins.GitHub.StandardRequirements.Private import Private
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "standard": {
            "private": True,
        },
    }


@pytest.fixture
def requirement():
    return Private()


class TestPrivate:
    def test_private_missing(self, requirement, query_data):
        """Test if `private` is missing"""
        query_data["standard"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_private_none(self, requirement, query_data):
        """Test if `private` is None"""
        query_data["standard"]["private"] = None
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_PublicRepo(self, requirement, query_data):
        """Test when repository is expected to be public."""
        query_data["standard"]["private"] = True
        requirement_args = {"enabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "The repository's visibility must be public." in result.context

    def test_Incorrect(self, requirement, query_data):
        """Test when repository is expected to be private but is actually not."""
        query_data["standard"]["private"] = False
        requirement_args = {"enabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "The repository's visibility must be private." in result.context

    def test_Successful(self, requirement, query_data):
        """Test when repository is private"""
        query_data["standard"]["private"] = True
        requirement_args = {"enabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success

        # Test when repository is public
        query_data["standard"]["private"] = False
        requirement_args = {"enabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success

        # Test using GetDynamicArgDefinitions
        query_data["standard"]["private"] = False
        requirement_args = {}
        for key, value in requirement.GetDynamicArgDefinitions().items():
            requirement_args[key] = value[1].default
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for DependabotSecurityUpdates.py"""

import pytest

from RepoAuditor.Plugins.GitHub.StandardRequirements.DependabotSecurityUpdates import (
    DependabotSecurityUpdates,
)
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "standard": {
            "security_and_analysis": {
                "dependabot_security_updates": {
                    "status": True,
                },
            },
        },
    }


@pytest.fixture
def requirement():
    return DependabotSecurityUpdates()


class TestDependabotSecurityUpdates:
    def test_security_and_analysis_missing(self, requirement, query_data):
        """Test if `security_and_analysis` is missing"""
        query_data["standard"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_dependabot_security_updates_missing(self, requirement, query_data):
        """Test if `dependabot_security_updates` is missing"""
        query_data["standard"]["security_and_analysis"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_status_missing(self, requirement, query_data):
        """Test if `status` is missing"""
        query_data["standard"]["security_and_analysis"]["dependabot_security_updates"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_status_none(self, requirement, query_data):
        """Test if `status` is None"""
        query_data["standard"]["security_and_analysis"]["dependabot_security_updates"]["status"] = None
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_Incorrect(self, requirement, query_data):
        """Result doesn't match expected value"""
        query_data["standard"]["security_and_analysis"]["dependabot_security_updates"]["status"] = True
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert (
            "'Dependabot security updates' must be set to 'True' (it is currently set to 'False')."
            in result.context
        )

    def test_Successful(self, requirement, query_data):
        """Successful"""
        query_data["standard"]["security_and_analysis"]["dependabot_security_updates"]["status"] = True
        requirement_args = {"disabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for Requirement.py"""

from unittest.mock import Mock

import pytest

from dbrownell_Common.Types import override

from RepoAuditor.Requirement import *


# ----------------------------------------------------------------------
class MyRequirement(Requirement):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        description: str,
        style: ExecutionStyle,
        resolution_template: str,
        rationale_template: str,
        expected_result: EvaluateResult,
        context: Optional[str] = None,
        *,
        requires_explicit_include: bool = False,
        provide_rationale: bool = True,
    ) -> None:
        super().__init__(
            name,
            description,
            style,
            resolution_template,
            rationale_template,
            requires_explicit_include=requires_explicit_include,
        )

        self.expected_result = expected_result
        self.context = context
        self._provide_rationale = provide_rationale

    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        return Requirement.EvaluateImplResult(
            self.expected_result,
            self.context,
            provide_resolution=self.expected_result == EvaluateResult.Error,
            provide_rationale=self._provide_rationale,
        )


# ----------------------------------------------------------------------
def test_Construct():
    name = Mock()
    description = Mock()
    style = Mock()
    resolution_template = Mock()
    rationale_template = Mock()
    expected_result = Mock()
    context = Mock()
    requires_explicit_include = Mock()

    r = MyRequirement(
        name,
        description,
        style,
        resolution_template,
        rationale_template,
        expected_result,
        context,
        requires_explicit_include=requires_explicit_include,
    )

    assert r.name is name
    assert r.description is description
    assert r.style is style
    assert r.resolution_template is resolution_template
    assert r.rationale_template is rationale_template
    assert r.expected_result is expected_result
    assert r.context is context
    assert r.requires_explicit_include is requires_explicit_include


# ----------------------------------------------------------------------
def test_Success():
    requirement = MyRequirement(Mock(), Mock(), Mock(), Mock(), Mock(), EvaluateResult.Success, "testing")
    result_info = requirement.Evaluate(Mock(), {})

    assert result_info.result == EvaluateResult.Success
    assert result_info.context == "testing"
    assert result_info.resolution is None
    assert result_info.rationale is None
    assert result_info.requirement is requirement


# ----------------------------------------------------------------------
def test_DoesNotApply():
    requirement = MyRequirement(Mock(), Mock(), Mock(), Mock(), Mock(), EvaluateResult.DoesNotApply, None)
    result_info = requirement.Evaluate(Mock(), {})

    assert result_info.result == EvaluateResult.DoesNotApply
    assert result_info.context is None
    assert result_info.resolution is None
    assert result_info.rationale is None
    assert result_info.requirement is requirement


# ----------------------------------------------------------------------
def test_GetDynamicArgDefinitions():
    requirement = MyRequirement(
        Mock(),
        Mock(),
        Mock(),
        Mock(),
        Mock(),
        EvaluateResult.Success,
        "testing",
    )

    assert requirement.GetDynamicArgDefinitions() == {}


# ----------------------------------------------------------------------
@pytest.mark.parametrize("provide_rationale", [True, False])
def test_Error(provide_rationale: bool):
    requirement = MyRequirement(
        Mock(),
        Mock(),
        Mock(),
        "{one} -- {two}",
        "{three} -- {four}",
        EvaluateResult.Error,
        "testing",
        provide_rationale=provide_rationale,
    )

    result_info = requirement.Evaluate({"one": "1", "two": "2", "three": "3", "four": "4"}, {"four": "FOUR"})

    assert result_info.result == EvaluateResult.Error
    assert result_info.context == "testing"
    assert result_info.resolution == "1 -- 2"
    assert result_info.rationale == ("3 -- 4" if provide_rationale else None)
    assert result_info.requirement is requirement


# ----------------------------------------------------------------------
def test_Warning():
    requirement = MyRequirement(
        Mock(),
        Mock(),
        Mock(),
        "{one} -- {two}",
        "{three} -- {four}",
        EvaluateResult.Warning,
        None,
    )

    result_info = requirement.Evaluate({"one": "1", "two": "2", "three": "3", "four": "4"}, {"four": "FOUR"})

    assert result_info.result == EvaluateResult.Warning
    assert result_info.context is None
    assert result_info.resolution is None
    assert result_info.rationale is None
    assert result_info.requirement is requirement


# ----------------------------------------------------------------------



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the EnableRequirementImpl object."""

from collections.abc import Callable
from typing import Any, Optional

import typer
from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.GitHub.Impl.Common import CreateIncompleteDataResult
from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement


# ----------------------------------------------------------------------
class EnableRequirementImpl(Requirement):
    """Object that implements settings that can be enabled/disabled."""

    # ----------------------------------------------------------------------
    def __init__(  # noqa: PLR0913
        self,
        name: str,
        enabled_by_default: bool,  # noqa: FBT001
        dynamic_arg_name: str,
        github_settings_value: str,
        get_configuration_value_func: Callable[[dict[str, Any]], Optional[bool]],
        resolution: str,
        rationale: str,
        subject: Optional[str] = None,
        *,
        requires_explicit_include: bool = False,
        unset_set_terminology: tuple[str, str] = ("unchecked", "checked"),
        missing_value_is_warning: bool = True,
    ) -> None:
        github_settings_value = f"'{github_settings_value}'"

        if subject is None:
            subject = github_settings_value

        super().__init__(
            name,
            f"Validates that {subject} is set to the expected value.",
            ExecutionStyle.Parallel,
            resolution,
            rationale,
            requires_explicit_include=requires_explicit_include,
        )

        self.dynamic_arg_name = dynamic_arg_name
        self.github_settings_value = github_settings_value
        self.enabled_by_default = enabled_by_default
        self.get_configuration_value_func = get_configuration_value_func
        self.unset_set_terminology = unset_set_terminology
        self.missing_value_is_warning = missing_value_is_warning

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Get the definitions for the arguments to this requirement."""
        return {
            self.dynamic_arg_name: (
                bool,
                typer.Option(
                    False,
                    help=f"Ensures that the check for {self.github_settings_value} is {'disabled' if self.enabled_by_default else 'enabled'}.",
                ),
            ),
        }

    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        result = self.get_configuration_value_func(query_data)
        if result is None:
            if query_data["session"].is_enterprise:
                return Requirement.EvaluateImplResult(
                    EvaluateResult.DoesNotApply,
                    (
                        "Please verify with your enterprise administrator "
                        f"that {self.github_settings_value} is enabled."
                    ),
                )

            if self.missing_value_is_warning:
                return CreateIncompleteDataResult()

            return Requirement.EvaluateImplResult(EvaluateResult.DoesNotApply, None)

        expected_value = self.enabled_by_default

        if requirement_args[self.dynamic_arg_name]:
            expected_value = not expected_value
            provide_rationale = False
        else:
            provide_rationale = True

        if result != expected_value:
            query_data["__checked_desc"] = self.unset_set_terminology[int(expected_value)]

            return Requirement.EvaluateImplResult(
                EvaluateResult.Error,
                f"{self.github_settings_value} must be set to '{expected_value}' (it is currently set to '{result}').",
                provide_resolution=True,
                provide_rationale=provide_rationale,
            )

        return Requirement.EvaluateImplResult(EvaluateResult.Success, None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the SquashCommitMerge object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class SquashCommitMerge(StandardEnableRequirementImpl):
    """Allow squash merging."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "SquashCommitMerge",
            False,
            "enabled",
            "settings",
            "Pull Requests",
            "Allow squash merging",
            lambda data: data["standard"].get("allow_squash_merge", None),
            textwrap.dedent(
                """\
                The default behavior is to not allow squash merging.

                Reasons for this Default
                ------------------------
                - When performing a Squash & Merge, GitHub creates a new merge commit with its key.
                This makes it difficult to verify author signatures when looking at the commit history.

                Reasons to Override this Default
                --------------------------------
                - Your repository does not require signatures.
                - You want to ensure that single-commit-changes are merged into the mainline branch to simplify the branch's history.
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the CommunityStandardsQuery object."""

from tempfile import TemporaryDirectory
from typing import Any, Optional
from urllib.parse import urlparse

from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Plugins.CommunityStandards.Requirements.CodeOfConduct import CodeOfConduct
from RepoAuditor.Plugins.CommunityStandards.Requirements.CodeOwners import CodeOwners
from RepoAuditor.Plugins.CommunityStandards.Requirements.Contributing import Contributing
from RepoAuditor.Plugins.CommunityStandards.Requirements.IssueTemplates import IssueTemplates
from RepoAuditor.Plugins.CommunityStandards.Requirements.LicenseFile import LicenseFile
from RepoAuditor.Plugins.CommunityStandards.Requirements.PullRequestTemplates import PullRequestTemplate
from RepoAuditor.Plugins.CommunityStandards.Requirements.ReadMe import ReadMe
from RepoAuditor.Plugins.CommunityStandards.Requirements.SecurityPolicy import SecurityPolicy
from RepoAuditor.Query import ExecutionStyle, Query


class CommunityStandardsQuery(Query):
    """Query with requirements that check for repository Community Standards files."""

    def __init__(self) -> None:
        super().__init__(
            "CommunityStandardsQuery",
            ExecutionStyle.Parallel,
            [
                ReadMe(),
                CodeOfConduct(),
                Contributing(),
                LicenseFile(),
                SecurityPolicy(),
                IssueTemplates(),
                PullRequestTemplate(),
                CodeOwners(),
            ],
        )

    # ----------------------------------------------------------------------
    @override
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        """Get the repo data."""

        # Clone the git repository to a temp directory
        branch = module_data.get("branch", "main")
        temp_repo_dir = TemporaryDirectory()
        url = module_data["url"]

        # Add PAT to git URL, if present
        pat = module_data.get("pat")
        if pat:
            parsed_url = urlparse(url)
            url = f"https://{pat}@{parsed_url.netloc}{parsed_url.path}"

        # Import git.Repo here so that it is only imported
        # if the CommunityStandards plugin is requested.
        from git import Repo

        try:
            Repo.clone_from(url, temp_repo_dir.name, branch=branch)
        except Exception as e:
            error_msg = f"""
            An error occurred while attempting to clone the target repository.
            If you are auditing a private repository, {
                "please ensure your PAT has access to the repository."
                if pat
                else "please provide a PAT with access to the repository."
            }
            Error: {e}
            """
            raise RuntimeError(error_msg) from e

        # Record the path and the temp directory for later cleanup
        module_data["repo_dir"] = temp_repo_dir

        return module_data

    @override
    def Cleanup(self, module_data: dict[str, Any]) -> None:
        """Clean up the cloned repository."""
        repo_dir = module_data["repo_dir"]
        repo_dir.cleanup()
        del module_data



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequirePullRequests object."""

from RepoAuditor.Plugins.GitHub.Impl.EnableRulesetRequirementImpl import EnableRulesetRequirementImpl


class RequirePullRequests(EnableRulesetRequirementImpl):
    """Require pull requests before merging to default branch."""

    def __init__(self) -> None:
        super().__init__(
            name="RequirePullRequestsRule",
            enabled_by_default=True,
            dynamic_arg_name="disabled",
            github_ruleset_type="pull_request",
            github_ruleset_value="Require a pull request before merging",
            get_configuration_value_func=lambda rule: rule.get("type", "") == "pull_request",
            resolution="{__expected_value} pull request rule in repository rulesets",
            rationale="Pull request reviews help maintain code quality and collaboration",
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains functionality to display the results of executing modules."""

import textwrap
from typing import Optional

from dbrownell_Common import TextwrapEx  # type: ignore[import-untyped]
from dbrownell_Common.Streams.DoneManager import DoneManager  # type: ignore[import-untyped]
from rich import print as rich_print
from rich.console import Group
from rich.panel import Panel

from RepoAuditor.Module import EvaluateResult, Module


def GetInternalPanelContent(
    dm: DoneManager,
    results: list[Module.EvaluateInfo],
    display_resolution: bool,  # noqa: FBT001
    display_rationale: bool,  # noqa: FBT001
    num_success: int,
    num_warning: int,
    num_error: int,
    num_does_not_apply: int,
    num_requirements: int,
) -> list[Panel | str]:
    """Get the content for each result to be displayed inside the internal panel."""
    internal_content: list[Panel | str] = []

    module = results[0].module

    if module.description:
        internal_content.append(f"{module.description.strip()}\n")

    metrics_display = textwrap.dedent(
        f"""\
            [grey66]
            Successful:     {num_success} ({(num_success / num_requirements):.02%})
            Warnings:       {num_warning} ({(num_warning / num_requirements):.02%})
            Errors:         {num_error} ({(num_error / num_requirements):.02%})
            Skipped:        {num_does_not_apply} ({(num_does_not_apply / num_requirements):.02%})
        """,
    )
    internal_content.append(metrics_display)

    for result in results:
        if not dm.is_verbose and result.result in [
            EvaluateResult.Success,
            EvaluateResult.DoesNotApply,
        ]:
            continue

        if result.result == EvaluateResult.Success:
            border_color = "green"
        elif result.result == EvaluateResult.Warning:
            border_color = "yellow"
        elif result.result == EvaluateResult.Error:
            border_color = "red"
        elif result.result == EvaluateResult.DoesNotApply:
            border_color = ""
        else:
            raise RuntimeError(result.result)  # pragma: no cover

        content: list[str] = []

        if result.requirement.description:
            content.append(f"{result.requirement.description.strip()}\n")

        if result.context:
            content.append(f"{result.result.name.upper()}: {result.context.strip()}\n")

        if display_resolution and result.resolution:
            content.append(
                textwrap.dedent(
                    """\
                        Resolution
                        ----------
                          {}
                        """,
                ).format(TextwrapEx.Indent(result.resolution.strip(), 2, skip_first_line=True)),
            )

        if display_rationale and result.rationale:
            content.append(
                textwrap.dedent(
                    """\
                        Rationale
                        ---------
                          {}
                        """,
                ).format(TextwrapEx.Indent(result.rationale.strip(), 2, skip_first_line=True)),
            )

        internal_content.append(
            Panel(
                Group(*content),
                title=f"[{result.result.name}] {result.requirement.name}",
                title_align="left",
                border_style=border_color,
                padding=(1, 1, 0, 1) if content else 0,
            ),
        )

    # Finally print statistics at the end in a panel if verbose
    internal_content.append(
        Panel(
            metrics_display,
            title="Metrics",
            title_align="left",
            border_style="grey66",
            padding=(1, 1, 0, 1),
        ),
    )

    return internal_content


# ----------------------------------------------------------------------
def DisplayResults(
    dm: DoneManager,
    all_results: list[list[Module.EvaluateInfo]],
    *,
    display_resolution: bool,
    display_rationale: bool,
    panel_width: Optional[int] = None,
) -> None:
    """Display the results of executing the modules."""
    for results in all_results:
        assert results
        module = results[0].module

        num_success = 0
        num_warning = 0
        num_error = 0
        num_does_not_apply = 0

        for result in results:
            if result.result == EvaluateResult.Success:
                num_success += 1
            elif result.result == EvaluateResult.Warning:
                num_warning += 1
            elif result.result == EvaluateResult.Error:
                num_error += 1
            elif result.result == EvaluateResult.DoesNotApply:
                num_does_not_apply += 1
            else:
                raise RuntimeError(result.result)  # pragma: no cover

        num_requirements = module.GetNumRequirements()
        assert num_requirements

        internal_content: list[Panel | str] = GetInternalPanelContent(
            dm,
            results,
            display_resolution,
            display_rationale,
            num_success=num_success,
            num_warning=num_warning,
            num_error=num_error,
            num_does_not_apply=num_does_not_apply,
            num_requirements=num_requirements,
        )

        rich_print(
            Panel(
                Group(*internal_content),
                padding=1 if internal_content else 0,
                title=results[0].module.name,
                width=panel_width,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for EntryPoint.py"""

import pytest
import typer

from dbrownell_Common.TestHelpers.StreamTestHelpers import InitializeStreamCapabilities
from dbrownell_Common.TyperEx import TypeDefinitionItem  # type: ignore [import-untyped]
from typer.testing import CliRunner

from RepoAuditor import __version__
from RepoAuditor.EntryPoint import app, TypeInfoToString

from .Plugins.utilities import GetGithubUrl


# ----------------------------------------------------------------------
@pytest.fixture(InitializeStreamCapabilities(), scope="session", autouse=True)
# ----------------------------------------------------------------------
def test_Version() -> None:
    result = CliRunner().invoke(app, ["--version"])

    assert result.exit_code == 0
    assert result.output == f"RepoAuditor v{__version__}\n"


# ----------------------------------------------------------------------
def test_Standard() -> None:
    result = CliRunner().invoke(app, [])

    assert result.exit_code == 1
    assert "There are no modules to process." in result.output


# ----------------------------------------------------------------------
@pytest.mark.skip(reason="This test isn't stable due to GitHub rate limiting for requests without a PAT.")
def test_GitHub() -> None:
    result = CliRunner().invoke(
        app,
        [
            "--include",
            "GitHub",
            "--GitHub-url",
            GetGithubUrl(),
        ],
    )

    assert result.exit_code == 1, result.output
    assert "Incomplete data was encountered; please provide the GitHub PAT." in result.output


# ----------------------------------------------------------------------
def test_Help() -> None:
    result = CliRunner().invoke(app, ["--help"])

    assert result.exit_code == 0
    assert "Module Information" in result.output


def test_TypeInfoToString() -> None:
    """Test the TypeInfoToString function to ensure valid command line output."""
    result = TypeInfoToString(
        "test_arg",
        (
            bool,
            typer.Option(
                False,
                help="Help text.",
            ),
        ),
    )
    expected_string = "    test_arg                                           bool    Help text."
    assert expected_string == result

    result = TypeInfoToString(
        "test_arg",
        TypeDefinitionItem(
            bool,
            typer.models.ParameterInfo(
                default=False,
                help="TypeDefinitionItem",
            ),
        ),
    )
    expected_string = "    test_arg                                           bool    TypeDefinitionItem"
    assert expected_string == result

    result = TypeInfoToString(
        "test_none",
        bool,
    )
    expected_string = "    test_none                                          bool    "
    assert expected_string == result



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the AllowDeletions object."""

import textwrap

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class AllowDeletions(ClassicEnableRequirementImpl):
    """Allow deletion of the mainline branch."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "AllowDeletions",
            False,
            "enabled",
            "Rules applied to everyone including administrators",
            "Allow deletions",
            lambda data: data["branch_protection_data"]["allow_deletions"]["enabled"],
            textwrap.dedent(
                """\
                The default behavior is to not allow the deletion of the mainline branch.

                Reasons for this Default
                ------------------------
                - Bad things happen when the mainline branch is deleted.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains Plugin functionality."""

import pluggy

from RepoAuditor import APP_NAME
from RepoAuditor.Module import Module
from RepoAuditor.Plugins.CommunityStandards.Module import CommunityStandardsModule


# ----------------------------------------------------------------------
@pluggy.HookimplMarker(APP_NAME)
def GetModule() -> Module:
    """Return Community Standards Module."""
    return CommunityStandardsModule()



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the CodeOfConduct requirement."""

import textwrap

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import ExistsRequirementImpl


class CodeOfConduct(ExistsRequirementImpl):
    """Validates that a CODE_OF_CONDUCT file is configured."""

    def __init__(self) -> None:
        super().__init__(
            "CodeOfConduct",
            "CODE_OF_CONDUCT",
            [
                ".github/CODE_OF_CONDUCT",
                ".github/CODE_OF_CONDUCT.md",
                ".github/CODE_OF_CONDUCT.rst",
                "docs/CODE_OF_CONDUCT",
                "docs/CODE_OF_CONDUCT.md",
                "docs/CODE_OF_CONDUCT.rst",
                "CODE_OF_CONDUCT",
                "CODE_OF_CONDUCT.md",
                "CODE_OF_CONDUCT.rst",
            ],
            textwrap.dedent(
                """\
                1) Create a CODE_OF_CONDUCT file in one of these locations:
                   - .github/CODE_OF_CONDUCT
                   - docs/CODE_OF_CONDUCT
                   - CODE_OF_CONDUCT

                Example CODE_OF_CONDUCT file format:
                ```
                # Contributor Code of Conduct
                ## Principles
                ## Community Guidelines
                ```
                """
            ),
            textwrap.dedent(
                """\
                A CODE_OF_CONDUCT defines expectations for behavior for your
                project's participants.

                Benefits:
                - Define community standards
                - Signal a welcoming and inclusive project
                - Outline procedures for handling abuse
                - Help create a positive social atmosphere for your community
                """
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Common test fixtures for full test suite"""

import os
import sys

import pytest


@pytest.fixture(autouse=True)
def fixed_terminal_size(monkeypatch):
    """Automatically set terminal to a fixed size for snapshot testing.
    This way the user's terminal size won't matter, leading to consistency.
    """
    monkeypatch.setattr(
        os,
        "get_terminal_size",
        lambda fd=sys.stdout.fileno(): os.terminal_size((120, 100)),
    )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Common test fixtures for Plugins"""

import pytest
from GitHubModule.fixtures import session

from RepoAuditor.Plugins.GitHubBase.Module import _GitHubSession


@pytest.fixture
def module_data():
    return {
        "session": _GitHubSession("https://github.com/gt-sse-center/RepoAuditor", "pat"),
        "url": "https://github.com/gt-sse-center/RepoAuditor",
    }



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireSignedCommits object."""

from RepoAuditor.Plugins.GitHub.Impl.EnableRulesetRequirementImpl import EnableRulesetRequirementImpl


class RequireSignedCommits(EnableRulesetRequirementImpl):
    """Check that the "Require signed commits" rule is disabled."""

    def __init__(self) -> None:
        super().__init__(
            name="RequireSignedCommitsRule",
            enabled_by_default=True,
            dynamic_arg_name="disabled",
            github_ruleset_type="required_signatures",
            github_ruleset_value="Require signed commits",
            get_configuration_value_func=lambda rule: rule.get("type", "") == "required_signatures",
            resolution="Enable commit signing requirement in repository rulesets",
            rationale="Signed commits ensure commit authenticity",
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the AllowMainlineForcePushes object."""

import textwrap

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class AllowMainlineForcePushes(ClassicEnableRequirementImpl):
    """Allow force pushes to the mainline branch."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "AllowMainlineForcePushes",
            False,
            "enabled",
            "Rules applied to everyone including administrators",
            "Allow force pushes",
            lambda data: data["branch_protection_data"]["allow_force_pushes"]["enabled"],
            textwrap.dedent(
                """\
                The default behavior is to not allow force pushes to the mainline branch.

                Reasons for this Default
                ------------------------
                - Force pushes rewrite history, which breaks git lineage for all other clones of the repository.
                  Merges are possible when this happens, but they are difficult to perform and data loss is
                  possible.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the StandardEnableRequirementImpl object."""

import textwrap
from collections.abc import Callable
from typing import Any, Optional

from RepoAuditor.Plugins.GitHub.Impl.EnableRequirementImpl import EnableRequirementImpl


# ----------------------------------------------------------------------
class StandardEnableRequirementImpl(EnableRequirementImpl):
    """Implementation of requirement for standard enables of GitHub settings."""

    # ----------------------------------------------------------------------
    def __init__(  # noqa: PLR0913
        self,
        name: str,
        enabled_by_default: bool,  # noqa: FBT001
        dynamic_arg_name: str,
        github_settings_url_suffix: str,
        github_settings_section: str,
        github_settings_value: str,
        get_configuration_value_func: Callable[[dict[str, Any]], Optional[bool]],
        rationale: str,
        subject: Optional[str] = None,
        *,
        requires_explicit_include: bool = False,
        unset_set_terminology: tuple[str, str] = ("unchecked", "checked"),
        missing_value_is_warning: bool = True,
    ) -> None:
        super().__init__(
            name,
            enabled_by_default,
            dynamic_arg_name,
            github_settings_value,
            get_configuration_value_func,
            textwrap.dedent(
                f"""\
                1) Visit '{{session.github_url}}/{github_settings_url_suffix}'
                2) Locate the '{github_settings_section}' section
                3) Ensure that '{github_settings_value}' is {{__checked_desc}}
                """,
            ),
            rationale,
            subject,
            requires_explicit_include=requires_explicit_include,
            unset_set_terminology=unset_set_terminology,
            missing_value_is_warning=missing_value_is_warning,
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the DismissStalePullRequestApprovals object."""

import textwrap
from typing import Any, Optional

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class DismissStalePullRequestApprovals(ClassicEnableRequirementImpl):
    """Dismiss stale pull request approvals when new commits are pushed."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "DismissStalePullRequestApprovals",
            True,
            "disabled",
            "Protect matching branches",
            "Require a pull request before merging -> Dismiss stale pull request approvals when new commits are pushed",
            _GetValue,
            textwrap.dedent(
                """\
                The default behavior is to dismiss stale pull request approvals when new commits are pushed.

                Reasons for this Default
                ------------------------
                - Approvals apply to the changes as they existed when the approval was granted. The approval
                  may no longer be valid when new changes are pushed.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """
            ),
        )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GetValue(
    data: dict[str, Any],
) -> Optional[bool]:
    settings = data["branch_protection_data"].get("required_pull_request_reviews", None)
    if settings is None:
        return None

    return settings.get("dismiss_stale_reviews", None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the Contributing requirement."""

import textwrap

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import ExistsRequirementImpl


class Contributing(ExistsRequirementImpl):
    """Validates that a contributing guide is configured."""

    def __init__(self) -> None:
        super().__init__(
            "Contributing",
            "CONTRIBUTING",
            [
                ".github/CONTRIBUTING.md",
                "docs/CONTRIBUTING.md",
                "CONTRIBUTING.md",
            ],
            textwrap.dedent(
                """\
                1) Create a contributing guide in one of these locations:
                   - .github/CONTRIBUTING.md
                   - docs/CONTRIBUTING.md
                   - CONTRIBUTING.md

                Example template can be found at:
                https://github.com/nayafia/contributing-template/blob/master/CONTRIBUTING-template.md
                """
            ),
            textwrap.dedent(
                """\
                Contributing guides help new contributors understand how to participate.

                Benefits:
                - Clear contribution process
                - Consistent code quality
                - Better community engagement
                """
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit test for Query.py"""

from unittest.mock import Mock

import pytest

from dbrownell_Common.Types import override

from RepoAuditor.Query import *


# ----------------------------------------------------------------------
def test_StatusInfo():
    status_info = StatusInfo()

    assert status_info.num_completed == 0
    assert status_info.num_success == 0
    assert status_info.num_error == 0
    assert status_info.num_warning == 0
    assert status_info.num_does_not_apply == 0

    # Fields must be in this exact order, as code relies on it
    # to easily call OnStatusFunc.
    assert list(StatusInfo().__dict__.keys()) == [
        "num_completed",
        "num_success",
        "num_error",
        "num_warning",
        "num_does_not_apply",
    ]


# ----------------------------------------------------------------------
class MyQuery(Query):
    # ----------------------------------------------------------------------
    @override
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        module_data["new_attribute"] = "NEW"

        return module_data


# ----------------------------------------------------------------------
class MyRequirement(Requirement):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        description: str,
        style: ExecutionStyle,
        resolution_template: str,
        rationale_template: str,
        expected_result: EvaluateResult,
        context: Optional[str] = None,
    ) -> None:
        super().__init__(name, description, style, resolution_template, rationale_template)

        self.expected_result = expected_result
        self.context = context

    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        return Requirement.EvaluateImplResult(
            self.expected_result,
            self.context,
            provide_resolution=True,
            provide_rationale=True,
        )


# ----------------------------------------------------------------------
requirement1 = MyRequirement(
    "MyRequirement1",
    "The first requirement",
    ExecutionStyle.Sequential,
    "{one} -- {two} -- {new_attribute}",
    "{three} -- {four} -- {new_attribute}",
    EvaluateResult.Success,
    None,
)

requirement2 = MyRequirement(
    "MyRequirement2",
    "The second requirement",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {new_attribute}",
    "{three} -- {four} -- {new_attribute}",
    EvaluateResult.Success,
    None,
)

requirement3 = MyRequirement(
    "MyRequirement3",
    "The third requirement",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {new_attribute}",
    "{three} -- {four} -- {new_attribute}",
    EvaluateResult.DoesNotApply,
    None,
)

requirement4 = MyRequirement(
    "MyRequirement4",
    "The fourth requirement",
    ExecutionStyle.Sequential,
    "{one} -- {two} -- {new_attribute}",
    "{three} -- {four} -- {new_attribute}",
    EvaluateResult.Error,
    None,
)

requirement5 = MyRequirement(
    "MyRequirement5",
    "The fifth requirement",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {new_attribute}",
    "{three} -- {four} -- {new_attribute}",
    EvaluateResult.Warning,
    None,
)

my_query = MyQuery(
    "MyQuery",
    ExecutionStyle.Sequential,
    [
        requirement1,
        requirement2,
        requirement3,
        requirement4,
        requirement5,
    ],
)


# ----------------------------------------------------------------------`
def test_MyQueryConstruct():
    assert my_query.name == "MyQuery"
    assert my_query.style == ExecutionStyle.Sequential
    assert my_query.requirements == [
        requirement1,
        requirement2,
        requirement3,
        requirement4,
        requirement5,
    ]


# ----------------------------------------------------------------------
@pytest.mark.parametrize("single_threaded", [True, False])
def test_Evaluate(single_threaded):
    module_data = {"one": 1, "two": 2, "three": 3, "four": 4}

    query_data = my_query.GetData(dict(module_data))
    assert query_data is not None

    status_func = Mock()

    results = my_query.Evaluate(
        query_data,
        {},
        status_func,
        max_num_threads=1 if single_threaded else None,
    )

    assert len(results) == 5

    assert results[0].result == EvaluateResult.Success
    assert results[0].query is my_query
    assert results[0].requirement is requirement1
    assert results[0].context is None
    assert results[0].resolution is None
    assert results[0].rationale is None

    assert results[1].result == EvaluateResult.Success
    assert results[1].query is my_query
    assert results[1].requirement is requirement2
    assert results[1].context is None
    assert results[1].resolution is None
    assert results[1].rationale is None

    assert results[2].result == EvaluateResult.DoesNotApply
    assert results[2].query is my_query
    assert results[2].requirement is requirement3
    assert results[2].context is None
    assert results[2].resolution is None
    assert results[2].rationale is None

    assert results[3].result == EvaluateResult.Error
    assert results[3].query is my_query
    assert results[3].requirement is requirement4
    assert results[3].context is None
    assert results[3].resolution == "1 -- 2 -- NEW"
    assert results[3].rationale == "3 -- 4 -- NEW"

    assert results[4].result == EvaluateResult.Warning
    assert results[4].query is my_query
    assert results[4].requirement is requirement5
    assert results[4].context is None
    assert results[4].resolution is None
    assert results[4].rationale is None

    if single_threaded:
        assert [call_args.args for call_args in status_func.call_args_list] == [
            (0, 0, 0, 0, 0),
            (1, 1, 0, 0, 0),
            (2, 1, 0, 0, 1),
            (3, 1, 0, 1, 1),
            (4, 2, 0, 1, 1),
            (5, 2, 1, 1, 1),
        ]



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for the GitHub Rulesets Plugin"""

from typing import Any

from RepoAuditor.Plugins.GitHub.RulesetRequirements.RequirePullRequests import RequirePullRequests
from RepoAuditor.Plugins.GitHub.RulesetRequirements.RequireSignedCommits import (
    RequireSignedCommits,
)
from RepoAuditor.Plugins.GitHub.RulesetRequirements.RequireStatusChecks import RequireStatusChecks
from RepoAuditor.Requirement import EvaluateResult


def create_rule(name: str, rule_type: str, ruleset_name: str, parameters: dict[str, Any]):
    """Helper function to create rule structure"""
    return {
        "name": name,
        "type": rule_type,
        "ruleset": {"name": ruleset_name},
        "parameters": parameters,
    }


# ----------------------------------------------------------------------
class TestRequirePullRequests:
    """Unit tests for the RequirePullRequests requirement."""

    def test_Enabled(self):
        """Test if the rule is enabled on the main branch."""
        requirement = RequirePullRequests()
        context = {
            "rules": [
                create_rule(
                    name="PR Rules",
                    rule_type="pull_request",
                    ruleset_name="Ruleset Main",
                    parameters={},
                )
            ]
        }
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Success
        assert f"{requirement.github_settings_value} is enabled" in result.context

    def test_Disabled(self):
        """Test if the required rule is disabled on the main branch
        by providing the wrong rule type.
        """
        requirement = RequirePullRequests()
        context = {
            "rules": [
                create_rule(
                    name="Inactive PR Rules",
                    # Wrong rule type
                    rule_type="push_request",
                    ruleset_name="Ruleset Main",
                    parameters={},
                )
            ]
        }
        result = requirement.Evaluate(context, {"disabled": True})
        assert result.result == EvaluateResult.Success

    def test_WrongTarget(self):
        """Test if rules are applied on the wrong branch."""
        requirement = RequirePullRequests()
        context = {
            "rules": [
                # GitHub behavior is to always give enforced rules on the specified target target.
                # Hence the rules here will be empty.
            ]
        }
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Error

    def test_NoRulesets(self):
        """Test if no rules are applied on the main branch."""
        requirement = RequirePullRequests()
        context = {"rules": []}
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Error
        assert f"No active branch ruleset with {requirement.github_settings_value} found" in result.context


# ----------------------------------------------------------------------
class TestRequireSignedCommits:
    """Unit tests for the RequireSignedCommits requirement."""

    def test_Enabled(self):
        """Test if the rule is enabled on the main branch."""
        requirement = RequireSignedCommits()
        context = {
            "rules": [
                create_rule(
                    name="Signing Rules",
                    rule_type="required_signatures",
                    ruleset_name="Ruleset Main",
                    parameters={},
                )
            ]
        }
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Success
        assert f"{requirement.github_settings_value} is enabled" in result.context

    def test_Disabled(self):
        """Test if the required rule is disabled on the main branch
        by providing the wrong rule type.
        """
        requirement = RequireSignedCommits()
        context = {
            "rules": [
                create_rule(
                    name="Signing Rules",
                    # Wrong rule type
                    rule_type="pull_request",
                    ruleset_name="Ruleset Main",
                    parameters={},
                )
            ]
        }
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Error

    def test_PartialConfig(self):
        """Test for rules that are not actively enforced."""
        requirement = RequireSignedCommits()
        context = {
            "rules": [
                # GitHub only returns rules that are actively enforced
            ]
        }
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Error

    def test_NoRulesets(self):
        """Test if no rules are applied on the main branch."""
        requirement = RequireSignedCommits()
        context = {"rules": []}
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Error
        assert f"No active branch ruleset with {requirement.github_settings_value} found" in result.context


# ----------------------------------------------------------------------
class TestRequireStatusChecks:
    """Unit tests for the RequireStatusChecks requirement."""

    def test_Enabled(self):
        """Test if the rule is enabled on the main branch."""
        requirement = RequireStatusChecks()
        context = {
            "rules": [
                create_rule(
                    name="CI Checks",
                    rule_type="required_status_checks",
                    ruleset_name="Ruleset Main",
                    parameters={},
                )
            ]
        }
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Success
        assert f"{requirement.github_settings_value} is enabled" in result.context

    def test_EmptyChecks(self):
        """Test for Status Check rules which don't have an associated CI check."""
        requirement = RequireStatusChecks()
        context = {
            "rules": [
                # GitHub won't allow active status check rules
                # which don't have associated CI checks.
                # Hence empty.
            ]
        }
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Error

    def test_NoRulesets(self):
        """Test if no rules are applied on the main branch."""
        requirement = RequireStatusChecks()
        context = {"rules": []}
        result = requirement.Evaluate(context, {"disabled": False})
        assert result.result == EvaluateResult.Error
        assert f"No active branch ruleset with {requirement.github_settings_value} found" in result.context



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for GitHub/Impl/ClassicValueRequirementImpl.py"""

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicValueRequirementImpl import (
    ClassicValueRequirementImpl,
)


class TestClassicValueRequirementImpl:
    def test_SettingsValue(self):
        """Test the ClassicValueRequirementImpl class."""
        requirement = ClassicValueRequirementImpl(
            "Require Some Value",
            "Disabled",
            github_settings_section=None,
            github_settings_value="No Settings Section",
            get_configuration_value_func=lambda _: "yes",
            rationale="For testing",
        )
        assert requirement.github_value == "'No Settings Section'"

    def test_NoSettingsValue(self):
        """Test for when no github_settings_value is provided."""
        requirement = ClassicValueRequirementImpl(
            "Require Some Value",
            "Disabled",
            github_settings_section="Protect funky feature",
            github_settings_value=None,
            get_configuration_value_func=lambda _: "yes",
            rationale="For testing",
        )
        assert requirement.github_value == "the entity"



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""End-to-end tests for the GitHub plugin"""

import os
from pathlib import Path

import pytest
from dbrownell_Common.TestHelpers.StreamTestHelpers import (
    InitializeStreamCapabilities,
)
from typer.testing import CliRunner
from utilities import CheckPATFileExists, GetGithubUrl, ScrubDurationGithuburlAndSpaces

from RepoAuditor.EntryPoint import app

# ----------------------------------------------------------------------
pytest.fixture(InitializeStreamCapabilities(), scope="session", autouse=True)


# ----------------------------------------------------------------------
def test_Successful(pat_args, snapshot):
    """Test if all requirements of the GitHub plugin successfully pass."""
    result = CliRunner().invoke(app, pat_args)

    assert result.exit_code == 0, result.output
    assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot


# ----------------------------------------------------------------------
def _skip_test_NoPAT():
    operating_system = os.getenv("GITHUB_CI_OPERATING_SYSTEM")
    python_version = os.getenv("GITHUB_CI_PYTHON_VERSION")

    if operating_system is None or python_version is None:
        # Always run locally
        return False

    # The selection of these values is arbitrary and serve only to ensure that the test is run on
    # one configuration.
    return operating_system != "ubuntu-latest" or python_version != "3.12"


@pytest.mark.skipif(
    _skip_test_NoPAT(),
    reason="Only run this test on the CI machine in specific configurations to avoid GitHub's request throttling.",
)
def test_NoPAT(snapshot, args):
    """Only run this test for specific CI conditions."""
    result = CliRunner().invoke(app, args)

    assert result.exit_code == 1, result.output
    assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot


# ----------------------------------------------------------------------
class TestStandard:
    # ----------------------------------------------------------------------
    def test_NoAutoMerge(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-AutoMerge-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoDeleteHeadBranches(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-DeleteHeadBranches-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoDependabotSecurityUpdates(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-DependabotSecurityUpdates-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoMergeCommit(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-MergeCommit-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesRebaseMergeCommit(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-RebaseMergeCommit-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoSecretScanning(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SecretScanning-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoSecretScanningPushProtection(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SecretScanningPushProtection-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesSquashCommitMerge(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SquashCommitMerge-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesSuggestUpdatingPullRequestBranches(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SuggestUpdatingPullRequestBranches-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesSupportDiscussions(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SupportDiscussions-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoSupportIssues(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SupportIssues-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoSupportProjects(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SupportProjects-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoSupportWikis(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-SupportWikis-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_YesTemplateRepository(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-TemplateRepository-enabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoWebCommitSignoff(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-WebCommitSignoff-disabled"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_DefaultBranchValue(self, pat_args, snapshot):
        result = CliRunner().invoke(app, pat_args + ["--GitHub-DefaultBranch-value", "not_main"])

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_NoDescription(self, pat_args, snapshot, monkeypatch):
        """Test if no repository description is available.
        The way this test works is that we use the default `--GitHub-Description-allow-empty=false`
        and monkey-patch the response to set `description` to empty.
        """
        from RepoAuditor.Plugins.GitHub.StandardQuery import StandardQuery

        def MockedGetData(
            self,
            module_data,
        ):
            """Monkey-patched class method to set description to empty."""
            response = module_data["session"].get("").json()
            response["description"] = ""
            module_data["standard"] = response
            return module_data

        monkeypatch.setattr(StandardQuery, "GetData", MockedGetData)

        result = CliRunner().invoke(app, pat_args)

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_LicenseValue(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-License-value", "Not the MIT License"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_MergeCommitMessageValue(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--GitHub-MergeCommitMessage-value",
                "This is not a valid merge commit message",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_PrivateValue(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-Private-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_SquashMergeCommitMessageValue(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args
            + [
                "--GitHub-SquashMergeCommitMessage-value",
                "This is not a valid squash merge commit message",
            ],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_Protected(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-Protected-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot


# ----------------------------------------------------------------------
class TestClassic:
    # ----------------------------------------------------------------------
    def test_AllowDeletions(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-AllowDeletions-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_AllowMainlineForcePush(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-AllowMainlineForcePushes-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_DismissStalePullRequestApprovals(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-DismissStalePullRequestApprovals-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_DoNotAllowBypassSettings(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-DoNotAllowBypassSettings-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_EnsureStatusChecks(self, pat_args):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-EnsureStatusChecks-disabled", "--verbose"],
        )

        assert result.exit_code == 0, result.output
        assert "[DoesNotApply] EnsureStatusChecks" in result.stdout

    # ----------------------------------------------------------------------
    def test_RequireApprovalMostRecentPush(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireApprovalMostRecentPush-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireApprovals(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireApprovals-value", "2"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireCodeOwnerReview(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireCodeOwnerReview-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireConversationResolution(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireConversationResolution-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireLinearHistory(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireLinearHistory-enabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequirePullRequests(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequirePullRequests-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireSignedCommits(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireSignedCommits-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireStatusChecksToPass(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireStatusChecksToPass-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    # ----------------------------------------------------------------------
    def test_RequireUpToDateBranches(self, pat_args, snapshot):
        result = CliRunner().invoke(
            app,
            pat_args + ["--GitHub-RequireUpToDateBranches-disabled"],
        )

        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot


# ----------------------------------------------------------------------
class TestRulesets:
    """End-to-end tests for enabled/disabled rules within a ruleset in a GitHub repository."""

    def test_Default(self, pat_args, snapshot, monkeypatch):
        """Test for default case when no rulesets are present on the branch."""

        from RepoAuditor.Plugins.GitHub.RulesetQuery import RulesetQuery

        def MockGetData(
            self,
            module_data,
        ):
            """Mock version for GetData for RulesetQuery"""
            # Specify empty ruleset data as the default
            module_data["rules"] = []

            return module_data

        monkeypatch.setattr(RulesetQuery, "GetData", MockGetData)

        result = CliRunner().invoke(app, pat_args)

        # Error since no enabled rulesets
        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    def test_RequireStatusChecks(self, pat_args, snapshot):
        """Test for requirement `Require status checks to pass` is disabled in the ruleset for the `main` branch."""
        result = CliRunner().invoke(app, pat_args + ["--GitHub-RequireStatusChecksRule-disabled"])

        # Since the repo ruleset has the requirement enabled,
        # we should get an error.
        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    def test_RequirePullRequests(self, pat_args, snapshot):
        """Test for requirement `Require a pull request before merging` is disabled in the ruleset for the `main` branch."""
        result = CliRunner().invoke(app, pat_args + ["--GitHub-RequirePullRequestsRule-disabled"])

        # Since the repo ruleset has the requirement disabled,
        # we should get an error.
        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot

    def test_RequireSignedCommits(self, pat_args, snapshot):
        """Test for requirement `Require signed commits` is disabled in the ruleset for the `main` branch."""
        result = CliRunner().invoke(app, pat_args + ["--GitHub-RequireSignedCommitsRule-disabled"])

        # Since the repo ruleset has the requirement disabled,
        # we should get an error.
        assert result.exit_code == -1, result.output
        assert ScrubDurationGithuburlAndSpaces(result.stdout) == snapshot


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
@pytest.fixture
def args() -> list[str]:
    """Common arguments for GitHub plugin and corresponding URL."""
    return [
        "--include",
        "GitHub",
        "--GitHub-url",
        GetGithubUrl(),
    ]


# ----------------------------------------------------------------------
@pytest.fixture
def pat_args(args) -> list[str]:
    github_pat_filename = (Path(__file__).parent / "github_pat.txt").resolve()
    CheckPATFileExists(github_pat_filename)

    return args + ["--GitHub-pat", str(github_pat_filename)]



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the Query object and types used in its definition."""

import threading
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Optional, Protocol

from RepoAuditor.Impl.ParallelSequentialProcessor import ParallelSequentialProcessor
from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement, ReturnCode


# ----------------------------------------------------------------------
class OnStatusFunc(Protocol):
    """Functional to report status information."""

    def __call__(  # noqa: D102
        self,
        num_completed: int,
        num_success: int,
        num_error: int,
        num_warning: int,
        num_does_not_apply: int,
    ) -> None: ...


# ----------------------------------------------------------------------
class Query(ABC):
    """A collection of Requirements that operate on a consistent set of data."""

    # ----------------------------------------------------------------------
    # |
    # |  Public Types
    # |
    # ----------------------------------------------------------------------
    @dataclass(frozen=True)
    class EvaluateInfo(Requirement.EvaluateInfo):
        """Information associated with evaluating a Query against a set of data."""

        query: "Query"

    # ----------------------------------------------------------------------
    # |
    # |  Public Methods
    # |
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        style: ExecutionStyle,
        requirements: list[Requirement],
    ) -> None:
        self.name = name
        self.style = style
        self.requirements = requirements

    # ----------------------------------------------------------------------
    @abstractmethod
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        """Return the data object augmented with information required by the Requirements associated with this Query."""

    # ----------------------------------------------------------------------
    def Evaluate(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
        status_func: OnStatusFunc,
        *,
        max_num_threads: Optional[int] = None,
    ) -> list["Query.EvaluateInfo"]:
        """Evaluate the Query given the query data and the data from the requirements."""
        status_info = StatusInfo()
        status_info_lock = threading.Lock()

        status_func(*status_info.__dict__.values())

        # ----------------------------------------------------------------------
        def EvaluateRequirement(
            requirement: Requirement,
        ) -> tuple[int, Query.EvaluateInfo]:
            result_info = requirement.Evaluate(
                query_data,
                requirement_args.get(requirement.name, {}),
            )
            return_code = ReturnCode.SUCCESS

            with status_info_lock:
                status_info.num_completed += 1

                if result_info.result == EvaluateResult.DoesNotApply:
                    status_info.num_does_not_apply += 1
                    return_code = ReturnCode.DOESNOTAPPLY
                elif result_info.result == EvaluateResult.Success:
                    status_info.num_success += 1
                elif result_info.result == EvaluateResult.Error:
                    status_info.num_error += 1
                    return_code = ReturnCode.ERROR
                elif result_info.result == EvaluateResult.Warning:
                    status_info.num_warning += 1
                    return_code = ReturnCode.WARNING
                else:
                    raise RuntimeError(result_info.result)  # pragma: no cover

                status_func(*status_info.__dict__.values())

            return (
                return_code,
                Query.EvaluateInfo(
                    **{
                        "query": self,
                        **result_info.__dict__,
                    },
                ),
            )

        # ----------------------------------------------------------------------

        return ParallelSequentialProcessor(
            self.requirements,
            EvaluateRequirement,
            max_num_threads=max_num_threads,
        )

    # ----------------------------------------------------------------------
    def Cleanup(
        self,
        module_data: dict[str, Any],
    ) -> None:
        """Clean up any resources created during execution."""
        del module_data


# ----------------------------------------------------------------------
class StatusInfo:
    """Class used to store information required by OnStatusFunc."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        self.num_completed: int = 0
        self.num_success: int = 0
        self.num_error: int = 0
        self.num_warning: int = 0
        self.num_does_not_apply: int = 0



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the MergeCommitMessage object."""

import textwrap
from typing import Any

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardValueRequirementImpl import (
    DoesNotApplyResult,
    StandardValueRequirementImpl,
)


# ----------------------------------------------------------------------
class MergeCommitMessage(StandardValueRequirementImpl):
    """Allow merge commit message requirement."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "MergeCommitMessage",
            "BLANK",
            "settings",
            "Pull Requests",
            "Allow merge commits -> Default...",
            _GetValue,
            textwrap.dedent(
                """\
                Available values on GitHub:

                Default message [PR_TITLE]
                    Pull Request number and head branch on the first line; pull request title on the third line

                Default to pull request title [BLANK]
                    Pull Request title and number on the first line.

                Default to pull request title and description [PR_BODY]
                    Pull Request title and number on the first line; pull request description starting on the third line.

            The default setting is "Default to pull request title [BLANK]".

            Reasons for this Default
            ------------------------
            - Reduce redundant information by only duplicating the title of the commit(s).
            - PR_TITLE includes the head branch name, which oftentimes is not relevant information to preserve over time.
            - PR_BODY duplicates the title and description of the commit(s).

            Reasons to Override this Default
            --------------------------------
            <unknown>
            """,
            ),
        )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GetValue(data: dict[str, Any]) -> str | DoesNotApplyResult | None:
    data = data["standard"]

    result = data.get("allow_merge_commit", None)
    if result is None:
        return None

    if not result:
        return DoesNotApplyResult("Merge commits are not enabled.")

    return data.get("merge_commit_message", None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the Module object and types used in its definition."""

import threading
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Optional

from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]

from RepoAuditor.Impl.ParallelSequentialProcessor import ParallelSequentialProcessor
from RepoAuditor.Query import EvaluateResult, ExecutionStyle, OnStatusFunc, Query, StatusInfo
from RepoAuditor.Requirement import ReturnCode


# ----------------------------------------------------------------------
class Module(ABC):
    """A collection of Queries that operate on a consistent set of data."""

    # ----------------------------------------------------------------------
    # |
    # |  Public Types
    # |
    # ----------------------------------------------------------------------
    @dataclass(frozen=True)
    class EvaluateInfo(Query.EvaluateInfo):
        """Information associated with evaluating a Module against a set of data."""

        module: "Module"

    # ----------------------------------------------------------------------
    # |
    # |  Public Methods
    # |
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        description: str,
        style: ExecutionStyle,
        queries: list[Query],
        *,
        requires_explicit_include: bool = False,  # If True, the module must be explicitly included on the command line
    ) -> None:
        self.name = name
        self.description = description
        self.style = style
        self.queries = queries
        self.requires_explicit_include = requires_explicit_include

    # ----------------------------------------------------------------------
    def GetNumRequirements(self) -> int:
        """Get the total number of requirements across all queries."""
        return sum(len(query.requirements) for query in self.queries)

    # ----------------------------------------------------------------------
    def ProcessRequirements(
        self,
        included_names: set[str],
        excluded_names: set[str],
    ) -> None:
        """Process the requirements to remove invalid requirements and queries."""
        query_index = 0

        while query_index < len(self.queries):
            query = self.queries[query_index]

            requirement_index = 0

            while requirement_index < len(query.requirements):
                requirement = query.requirements[requirement_index]

                if (requirement.requires_explicit_include and requirement.name not in included_names) or (
                    requirement.name in excluded_names
                ):
                    query.requirements.pop(requirement_index)
                    continue

                requirement_index += 1

            if not query.requirements:
                self.queries.pop(query_index)
                continue

            query_index += 1

    # ----------------------------------------------------------------------
    @abstractmethod
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Return information about dynamic arguments that the module can consume (often from the command line)."""

        raise NotImplementedError("Abstract method")  # pragma: no cover # noqa: EM101

    # ----------------------------------------------------------------------
    @abstractmethod
    def GenerateInitialData(
        self,
        dynamic_args: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        """Augments data beyond data initially encountered (for example, on the command line).

        Return None to indicate that the Module cannot be evaluated. Raise an
        exception to indicate that the command line data is invalid.
        """

        raise NotImplementedError("Abstract method")  # pragma: no cover # noqa: EM101

    # ----------------------------------------------------------------------
    def Evaluate(
        self,
        module_data: dict[str, Any],
        requirement_data: dict[str, Any],
        status_func: OnStatusFunc,
        *,
        max_num_threads: Optional[int] = None,
    ) -> list[list["Module.EvaluateInfo"]]:
        """Evaluate the module using the module data and requirement data."""
        status_info = StatusInfo()
        status_info_lock = threading.Lock()

        # ----------------------------------------------------------------------
        def EvaluateQuery(
            query: Query,
        ) -> tuple[int, list[Module.EvaluateInfo]]:
            query_data = query.GetData(dict(module_data))
            if query_data is None:
                with status_info_lock:
                    status_info.num_completed += len(query.requirements)
                    status_info.num_does_not_apply += len(query.requirements)

                    status_func(*status_info.__dict__.values())
                    # Since query returned None, it means it was not valid.
                    return 2, [
                        Module.EvaluateInfo(
                            result=EvaluateResult.DoesNotApply,
                            context=f"{query.name} did not return valid data.",
                            resolution="",
                            rationale="",
                            requirement=requirement,
                            query=query,
                            module=self,
                        )
                        for requirement in query.requirements
                    ]

            prev_query_status_info = StatusInfo()

            # ----------------------------------------------------------------------
            def OnQueryStatus(
                num_completed: int,
                num_success: int,
                num_error: int,
                num_warning: int,
                num_does_not_apply: int,
            ) -> None:
                with status_info_lock:
                    status_info.num_completed += num_completed - prev_query_status_info.num_completed
                    status_info.num_success += num_success - prev_query_status_info.num_success
                    status_info.num_error += num_error - prev_query_status_info.num_error
                    status_info.num_warning += num_warning - prev_query_status_info.num_warning
                    status_info.num_does_not_apply += (
                        num_does_not_apply - prev_query_status_info.num_does_not_apply
                    )

                    status_func(*status_info.__dict__.values())

                prev_query_status_info.num_completed = num_completed
                prev_query_status_info.num_success = num_success
                prev_query_status_info.num_error = num_error
                prev_query_status_info.num_warning = num_warning
                prev_query_status_info.num_does_not_apply = num_does_not_apply

            # ----------------------------------------------------------------------

            evaluate_infos = query.Evaluate(
                query_data,
                requirement_data,
                OnQueryStatus,
                max_num_threads=max_num_threads,
            )

            return_code = ReturnCode.SUCCESS

            for evaluate_info in evaluate_infos:
                if evaluate_info.result == EvaluateResult.Error:
                    return_code = ReturnCode.ERROR
                    break

                if evaluate_info.result == EvaluateResult.Warning:
                    return_code = ReturnCode.WARNING

                if evaluate_info.result == EvaluateResult.DoesNotApply:
                    return_code = ReturnCode.DOESNOTAPPLY

            # Cleanup any resources created during the query
            query.Cleanup(query_data)

            return (
                return_code,
                [
                    Module.EvaluateInfo(
                        **{
                            "module": self,
                            **evaluate_info.__dict__,
                        },
                    )
                    for evaluate_info in evaluate_infos
                ],
            )

        # ----------------------------------------------------------------------

        return ParallelSequentialProcessor(
            self.queries,
            EvaluateQuery,
            max_num_threads=max_num_threads,
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains functionality to process items in parallel and/or sequentially."""

from collections.abc import Callable
from enum import Enum, auto
from typing import Optional, TypeVar, Union, cast

from dbrownell_Common import ExecuteTasks  # type: ignore[import-untyped]
from dbrownell_Common.Streams.DoneManager import DoneManager  # type: ignore[import-untyped]
from dbrownell_Common.Streams.StreamDecorator import StreamDecorator  # type: ignore[import-untyped]


# ----------------------------------------------------------------------
class ExecutionStyle(Enum):
    """Controls the way in which a Requirement, Query, and Module can be processed."""

    Sequential = auto()
    Parallel = auto()


# ----------------------------------------------------------------------
ItemType = TypeVar("ItemType")
OutputType = TypeVar("OutputType")


def ParallelSequentialProcessor(
    items: list[ItemType],
    calculate_result_func: Callable[[ItemType], tuple[int, OutputType]],
    dm: Optional[DoneManager] = None,
    *,
    max_num_threads: Optional[int] = None,
) -> list[OutputType]:
    """Process a list of items in parallel and/or sequentially."""
    if dm is None:
        with DoneManager.Create(StreamDecorator(None), "", line_prefix="") as _dm:
            return _Impl(
                _dm,
                items,
                calculate_result_func,
                max_num_threads,
            )

    return _Impl(
        dm,
        items,
        calculate_result_func,
        max_num_threads,
    )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _Impl(
    dm: DoneManager,
    items: list[ItemType],
    calculate_result_func: Callable[[ItemType], tuple[int, OutputType]],
    max_num_threads: Optional[int],
) -> list[OutputType]:
    # Divide the items into those that can be run in parallel and those that must be run sequentially
    parallel: list[tuple[int, ItemType]] = []
    sequential: list[tuple[int, ItemType]] = []

    for index, item in enumerate(items):
        execution_style = item.style  # type: ignore[attr-defined]

        if execution_style == ExecutionStyle.Parallel:
            parallel.append((index, item))
        elif execution_style == ExecutionStyle.Sequential:
            sequential.append((index, item))
        else:
            raise RuntimeError(execution_style)  # pragma: no cover

    if len(parallel) == 1:
        sequential.append(parallel[0])
        parallel = []

    # Calculate the results
    results: list[Optional[OutputType]] = [None] * len(items)

    # ----------------------------------------------------------------------
    def Execute(
        results_index: int,
        item: ItemType,
    ) -> ExecuteTasks.CompleteTransformResult:
        return_code, result = calculate_result_func(item)

        assert results[results_index] is None
        results[results_index] = result

        return ExecuteTasks.CompleteTransformResult(None, return_code)

    # ----------------------------------------------------------------------

    if parallel:
        transform_results: list[Union[None, object, Exception]] = ExecuteTasks.TransformTasks(
            dm,
            "Processing",
            [
                ExecuteTasks.TaskData(item.name, (results_index, item))  # type: ignore[attr-defined]
                for results_index, item in parallel
            ],
            lambda context, status: Execute(*context),  # noqa: ARG005
            max_num_threads=max_num_threads,
            return_exceptions=True,
        )

        for transform_result in transform_results:
            if transform_result is not None and isinstance(transform_result, Exception):
                raise transform_result  # pragma: no cover

    for sequential_index, (results_index, item) in enumerate(sequential):
        with dm.Nested(
            "Processing '{}' ({} of {})...".format(
                item.name,  # type: ignore[attr-defined]
                sequential_index + 1 + len(parallel),
                len(items),
            ),
        ):
            Execute(results_index, item)

    assert not any(result is None for result in results), results
    return cast(list[OutputType], results)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for GitHub/Impl/ClassicValueRequirementImpl.py"""

from pathlib import Path
from tempfile import TemporaryDirectory

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import ExistsRequirementImpl
from RepoAuditor.Requirement import EvaluateResult


class MockDirectory:
    def __init__(self):
        self.name = "repository"


class TestExistsRequirementImpl:
    def test_Constructor(self):
        """Test the requirement implementation constructor."""
        requirement = ExistsRequirementImpl(
            name="Exists Some Value",
            filename="README.md",
            possible_locations=[
                "README.md",
            ],
            resolution="Get test to pass",
            rationale="For testing",
        )
        assert requirement.filename == "README.md"

    def test_Disabled(self):
        """Test disabled requirement."""
        requirement = ExistsRequirementImpl(
            name="Exists Some Value",
            filename="README.md",
            possible_locations=[
                "README.md",
            ],
            resolution="Get test to pass",
            rationale="For testing",
        )

        query_data = {}
        requirement_args = {"unrequired": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply

    def test_NotFound(self):
        """Test for when file is not found."""
        requirement = ExistsRequirementImpl(
            name="Exists Some Value",
            filename="README.md",
            possible_locations=[
                "MEREAD.md",
            ],
            resolution="Get test to pass",
            rationale="For testing",
        )

        query_data = {"repo_dir": MockDirectory()}
        requirement_args = {"unrequired": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "No README.md file found" in result.context

    def test_Found(self):
        """Test for when file is found as is."""
        requirement = ExistsRequirementImpl(
            name="Exists Some Value",
            filename="README.md",
            possible_locations=[
                "README.md",
            ],
            resolution="Get test to pass",
            rationale="For testing",
        )

        tempdir = TemporaryDirectory()
        with open(Path(tempdir.name, "README.md"), "w") as f:
            f.write("temp file in temp directory")

        query_data = {"repo_dir": tempdir}
        requirement_args = {"unrequired": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success
        assert "README.md found in repository" in result.context

    def test_FoundInDirectory(self):
        """Test if provided path is directory."""
        requirement = ExistsRequirementImpl(
            name="Exists Some Value",
            filename="README.md",
            possible_locations=[
                "docs",
            ],
            resolution="Get test to pass",
            rationale="For testing",
        )

        tempdir = TemporaryDirectory()
        # Create "docs" directory
        docs_dir = Path(tempdir.name) / "docs"
        docs_dir.mkdir(parents=True)

        with open(docs_dir / "README.md", "w") as f:
            f.write("temp file in temp directory")

        query_data = {"repo_dir": tempdir}
        requirement_args = {"unrequired": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success
        assert "File found in docs directory of the repository" in result.context



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the interface that Plugins must implement."""

import pluggy

from RepoAuditor import APP_NAME
from RepoAuditor.Module import Module


# ----------------------------------------------------------------------
@pluggy.HookspecMarker(APP_NAME)
def GetModule() -> Module:
    """Return a Module."""
    raise NotImplementedError("hookspec")  # pragma: no cover # noqa: EM101



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireStatusChecksToPass object."""

import textwrap

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicEnableRequirementImpl import (
    ClassicEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class RequireStatusChecksToPass(ClassicEnableRequirementImpl):
    """Require status checks to pass before merging pull requests."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequireStatusChecksToPass",
            True,
            "disabled",
            "Protect matching branches",
            "Require status checks to pass before merging",
            lambda data: "required_status_checks" in data["branch_protection_data"],
            textwrap.dedent(
                """\
                The default behavior is to require status checks to pass before merging a pull request.

                Reasons for this Default
                ------------------------
                - Status checks are an important part of the development process and should not be bypassed.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for EnsureStatusChecks.py"""

import pytest

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.EnsureStatusChecks import (
    EnsureStatusChecks,
)
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "branch": "main",
        "branch_protection_data": {
            "required_status_checks": {
                "checks": ["check1", "check2"],
            },
        },
    }


@pytest.fixture
def requirement():
    return EnsureStatusChecks()


class TestEnsureStatusChecks:
    def test_disabled(self, requirement, query_data):
        """Test disabled requirement"""
        requirement_args = {"disabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert "The status check requirement has been explicitly disabled." in result.context

    def test_required_status_checks_missing(self, requirement, query_data):
        """Test when `required_status_checks` is missing"""
        query_data["branch_protection_data"] = {}
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_required_status_checks_none(self, requirement, query_data):
        """Test when `required_status_checks` is None"""
        query_data["branch_protection_data"] = {"required_status_checks": None}
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.DoesNotApply
        assert result.context is None

    def test_checks_missing(self, requirement, query_data):
        """Test when `checks` is missing"""
        query_data["branch_protection_data"]["required_status_checks"] = {}
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "No status checks are configured." in result.context

    def test_checks_none(self, requirement, query_data):
        """Test when `checks` is None"""
        query_data["branch_protection_data"]["required_status_checks"] = {"checks": None}
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "No status checks are configured." in result.context

    def test_Successful(self, requirement, query_data):
        """Test successful"""
        query_data["branch_protection_data"]["required_status_checks"]["checks"] = ["check1", "check2"]
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success

    def test_SuccessfulGetDynamicArgDefinitions(self, requirement, query_data):
        """Test with GetDynamicArgDefinitions"""
        query_data["branch_protection_data"]["required_status_checks"]["checks"] = ["check1", "check2"]
        requirement_args = {}
        for key, value in requirement.GetDynamicArgDefinitions().items():
            requirement_args[key] = value[1].default
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireApprovals object."""

import textwrap
from typing import Any

from RepoAuditor.Plugins.GitHub.ClassicBranchProtectionRequirements.Impl.ClassicValueRequirementImpl import (
    ClassicValueRequirementImpl,
    DoesNotApplyResult,
)


# ----------------------------------------------------------------------
class RequireApprovals(ClassicValueRequirementImpl):
    """Require approvals for pull requests."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "RequireApprovals",
            "1",
            "Protect matching branches",
            "Require a pull request before merging -> Require approvals",
            _GetValue,
            textwrap.dedent(
                """\
                The default behavior is to require at least one approval.

                Reasons for this Default
                ------------------------
                - Code Reviews are a generally accepted best practice for software development.

                Reasons to Override this Default
                --------------------------------
                - You are the only person working on the repository (set the value to 0).
                - You want to require more than one approval (set the value to the number of approvals required).
                - You are not comfortable with the amount of time code reviews introduce in the development process (set the value to 0).
                """,
            ),
        )


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _GetValue(
    data: dict[str, Any],
) -> str | DoesNotApplyResult | None:
    settings = data["branch_protection_data"].get("required_pull_request_reviews", None)
    if settings is None:
        return None

    return settings.get("required_approving_review_count", None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the ReadMe requirement."""

import textwrap

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import ExistsRequirementImpl


class ReadMe(ExistsRequirementImpl):
    """Validates that a README file is configured."""

    def __init__(self) -> None:
        super().__init__(
            "ReadMe",
            "README",
            [
                "README.md",
            ],
            textwrap.dedent(
                """\
                Create a README.md file in the root of the repository.

                Example READMEs can be found at https://github.com/matiassingers/awesome-readme
                """
            ),
            textwrap.dedent(
                """\
                A README file provides the first point of entry into using
                and contributing to the repository.

                Benefits:
                - Summarizes repository.
                - Instructions on how to install and run the code.
                - Provides links to more information.
                """
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the RequireStatusChecks object."""

from RepoAuditor.Plugins.GitHub.Impl.EnableRulesetRequirementImpl import EnableRulesetRequirementImpl


class RequireStatusChecks(EnableRulesetRequirementImpl):
    """Require status checks to pass before merging."""

    def __init__(self) -> None:
        super().__init__(
            name="RequireStatusChecksRule",
            enabled_by_default=True,
            dynamic_arg_name="disabled",
            github_ruleset_type="required_status_checks",
            github_ruleset_value="Require status checks to pass",
            get_configuration_value_func=lambda rule: rule.get("type", "") == "required_status_checks",
            resolution="Configure required status checks in repository rulesets",
            rationale="Status checks ensure code quality and compatibility",
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for SecretScanningPushProtection.py"""

import pytest

from RepoAuditor.Plugins.GitHub.StandardRequirements.SecretScanningPushProtection import (
    SecretScanningPushProtection,
)
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "standard": {
            "security_and_analysis": {
                "secret_scanning_push_protection": {
                    "status": True,
                },
            },
        },
    }


@pytest.fixture
def requirement():
    return SecretScanningPushProtection()


class TestSecretScanningPushProtection:
    def test_security_and_analysis_missing(self, requirement, query_data):
        """Test if `security_and_analysis` is missing"""
        query_data["standard"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_secret_scanning_push_protection_missing(self, requirement, query_data):
        """Test if `secret_scanning_push_protection` is missing"""
        query_data["standard"]["security_and_analysis"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_status_missing(self, requirement, query_data):
        """Test if `status` is missing"""
        query_data["standard"]["security_and_analysis"]["secret_scanning_push_protection"] = {}
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_status_none(self, requirement, query_data):
        """Test if `status` is None"""
        query_data["standard"]["security_and_analysis"]["secret_scanning_push_protection"]["status"] = None
        result = requirement.Evaluate(query_data, requirement_args={})
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_Incorrect(self, requirement, query_data):
        """Test when result doesn't match expected value"""
        query_data["standard"]["security_and_analysis"]["secret_scanning_push_protection"]["status"] = True
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert "'Push protection' must be set to 'True' (it is currently set to 'False')." in result.context

    def test_Successful(self, requirement, query_data):
        """Successful"""
        query_data["standard"]["security_and_analysis"]["secret_scanning_push_protection"]["status"] = True
        requirement_args = {"disabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the AutoMerge object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardEnableRequirementImpl import (
    StandardEnableRequirementImpl,
)


# ----------------------------------------------------------------------
class AutoMerge(StandardEnableRequirementImpl):
    """Requirement to enable auto-merge for pull requests."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "AutoMerge",
            True,
            "disabled",
            "settings",
            "Pull Requests",
            "Allow auto-merge",
            lambda data: data["standard"].get("allow_auto_merge", None),
            textwrap.dedent(
                """\
                The default behavior is to enable the option to auto-merge once all the required status checks associated with a pull request have passed.

                Reasons for this Default
                ------------------------
                - Reduces mean resolution time by triggering the merge once all the required status checks pass.

                Reasons to Override this Default
                --------------------------------
                <unknown>
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit test for Module.py"""

import copy
from typing import Optional, cast
from unittest.mock import Mock

import pytest
from dbrownell_Common.Streams.DoneManager import DoneManager
from dbrownell_Common.TestHelpers.StreamTestHelpers import GenerateDoneManagerAndContent
from dbrownell_Common.Types import override

from RepoAuditor.Module import *
from RepoAuditor.Requirement import EvaluateResult, Requirement


# ----------------------------------------------------------------------
class MyModule(Module):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        description: str,
        style: ExecutionStyle,
        queries: list[Query],
        *,
        produce_data: bool,
        requires_explicit_include: bool = False,
    ) -> None:
        super().__init__(
            name,
            description,
            style,
            queries,
            requires_explicit_include=requires_explicit_include,
        )

        self.produce_data = produce_data

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        return {}

    # ----------------------------------------------------------------------
    @override
    def GenerateInitialData(
        self,
        dynamic_args: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        if self.produce_data is False:
            return None

        assert not dynamic_args

        dynamic_args["produce_data"] = self.produce_data
        dynamic_args["one"] = 1
        dynamic_args["two"] = 2
        dynamic_args["three"] = 3
        dynamic_args["four"] = 4
        dynamic_args["module_data"] = self.name

        return dynamic_args


# ----------------------------------------------------------------------
class MyQuery(Query):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        style: ExecutionStyle,
        requirements: list[Requirement],
        *,
        produce_data: bool,
    ) -> None:
        super().__init__(name, style, requirements)

        self.produce_data = produce_data

    # ----------------------------------------------------------------------
    def GetData(
        self,
        module_data: dict[str, Any],
    ) -> Optional[dict[str, Any]]:
        if not self.produce_data:
            return None

        module_data["query_data"] = self.name

        return module_data


# ----------------------------------------------------------------------
class MyRequirement(Requirement):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        description: str,
        style: ExecutionStyle,
        resolution_template: str,
        rationale_template: str,
        evaluate_result: EvaluateResult,
        *,
        requires_explicit_include: bool = False,
    ) -> None:
        super().__init__(
            name,
            description,
            style,
            resolution_template,
            rationale_template,
            requires_explicit_include=requires_explicit_include,
        )

        self.evaluate_result = evaluate_result

    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        return Requirement.EvaluateImplResult(
            self.evaluate_result,
            None,
            provide_resolution=True,
            provide_rationale=True,
        )


# ----------------------------------------------------------------------
def test_Construct():
    name = Mock()
    description = Mock()
    style = Mock()
    queries = Mock()
    produce_data = Mock()
    requires_explicit_include = Mock()

    module = MyModule(
        name,
        description,
        style,
        queries,
        produce_data=produce_data,
        requires_explicit_include=requires_explicit_include,
    )

    assert module.name is name
    assert module.description is description
    assert module.style is style
    assert module.queries is queries
    assert module.produce_data is produce_data
    assert module.requires_explicit_include is requires_explicit_include


# ----------------------------------------------------------------------
requirementA1 = MyRequirement(
    "RequirementA1",
    "The first requirement for A",
    ExecutionStyle.Sequential,
    "{one} -- {two} -- {query_data} -- {module_data}",
    "{three} -- {four} -- {query_data} -- {module_data}",
    EvaluateResult.Success,
    requires_explicit_include=True,
)

requirementA2 = MyRequirement(
    "RequirementA2",
    "The second requirement for A",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {query_data} -- {module_data}",
    "{three} -- {four} -- {query_data} -- {module_data}",
    EvaluateResult.Success,
)

requirementA3 = MyRequirement(
    "RequirementA3",
    "The third requirement for A",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {query_data} -- {module_data}",
    "{three} -- {four} -- {query_data} -- {module_data}",
    EvaluateResult.Error,
)

requirementA4 = MyRequirement(
    "RequirementA4",
    "The fourth requirement for A",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {query_data} -- {module_data}",
    "{three} -- {four} -- {query_data} -- {module_data}",
    EvaluateResult.Success,
)

requirementB1 = MyRequirement(
    "RequirementB1",
    "The first requirement for B",
    ExecutionStyle.Sequential,
    "{one} -- {two} -- {query_data} -- {module_data}",
    "{three} -- {four} -- {query_data} -- {module_data}",
    EvaluateResult.Success,
    requires_explicit_include=True,
)

requirementB2 = MyRequirement(
    "RequirementB2",
    "The second requirement for B",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {query_data} -- {module_data}",
    "{three} -- {four} -- {query_data} -- {module_data}",
    EvaluateResult.Success,
)

requirementB3 = MyRequirement(
    "RequirementB3",
    "The third requirement for B",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {query_data} -- {module_data}",
    "{three} -- {four} -- {query_data} -- {module_data}",
    EvaluateResult.Success,
)

requirementB4 = MyRequirement(
    "RequirementB4",
    "The fourth requirement for B",
    ExecutionStyle.Parallel,
    "{one} -- {two} -- {query_data} -- {module_data}",
    "{three} -- {four} -- {query_data} -- {module_data}",
    EvaluateResult.Success,
)

queryA = MyQuery(
    "QueryA",
    ExecutionStyle.Parallel,
    [requirementA1, requirementA2, requirementA3, requirementA4],
    produce_data=True,
)

queryB = MyQuery(
    "QueryB",
    ExecutionStyle.Parallel,
    [requirementB1, requirementB2, requirementB3, requirementB4],
    produce_data=True,
)

queryC = MyQuery(
    "QueryC",
    ExecutionStyle.Sequential,
    [requirementA1, requirementA2, requirementA3],
    produce_data=False,
)


# ----------------------------------------------------------------------
@pytest.mark.parametrize("single_threaded", [True, False])
def test_Module(single_threaded):
    module = MyModule(
        "MyModule",
        "The only module",
        ExecutionStyle.Sequential,
        [queryA, queryB, queryC],
        produce_data=True,
    )

    status_func = Mock()

    initial_data = module.GenerateInitialData({})
    assert initial_data is not None

    results = module.Evaluate(
        initial_data,
        {},
        status_func,
        max_num_threads=1 if single_threaded else None,
    )

    assert len(results) == 3

    # QueryA
    assert len(results[0]) == 4

    assert results[0][0].result == EvaluateResult.Success
    assert results[0][0].context is None
    assert results[0][0].resolution is None
    assert results[0][0].rationale is None
    assert results[0][0].requirement is requirementA1
    assert results[0][0].query is queryA
    assert results[0][0].module is module

    assert results[0][1].result == EvaluateResult.Success
    assert results[0][1].context is None
    assert results[0][1].resolution is None
    assert results[0][1].rationale is None
    assert results[0][1].requirement is requirementA2
    assert results[0][1].query is queryA
    assert results[0][1].module is module

    assert results[0][2].result == EvaluateResult.Error
    assert results[0][2].context is None
    assert results[0][2].resolution == "1 -- 2 -- QueryA -- MyModule"
    assert results[0][2].rationale == "3 -- 4 -- QueryA -- MyModule"
    assert results[0][2].requirement is requirementA3
    assert results[0][2].query is queryA
    assert results[0][2].module is module

    assert results[0][3].result == EvaluateResult.Success
    assert results[0][3].context is None
    assert results[0][3].resolution is None
    assert results[0][3].rationale is None
    assert results[0][3].requirement is requirementA4
    assert results[0][3].query is queryA
    assert results[0][3].module is module

    # QueryB
    assert len(results[1]) == 4

    assert results[1][0].result == EvaluateResult.Success
    assert results[1][0].context is None
    assert results[1][0].resolution is None
    assert results[1][0].rationale is None
    assert results[1][0].requirement is requirementB1
    assert results[1][0].query is queryB
    assert results[1][0].module is module

    assert results[1][1].result == EvaluateResult.Success
    assert results[1][1].context is None
    assert results[1][1].resolution is None
    assert results[1][1].rationale is None
    assert results[1][1].requirement is requirementB2
    assert results[1][1].query is queryB
    assert results[1][1].module is module

    assert results[1][2].result == EvaluateResult.Success
    assert results[1][2].context is None
    assert results[1][2].resolution is None
    assert results[1][2].rationale is None
    assert results[1][2].requirement is requirementB3
    assert results[1][2].query is queryB
    assert results[1][2].module is module

    assert results[1][3].result == EvaluateResult.Success
    assert results[1][3].context is None
    assert results[1][3].resolution is None
    assert results[1][3].rationale is None
    assert results[1][3].requirement is requirementB4
    assert results[1][3].query is queryB
    assert results[1][3].module is module

    # QueryC
    query_c_expected_results = [
        Module.EvaluateInfo(
            result=EvaluateResult.DoesNotApply,
            context="QueryC did not return valid data.",
            rationale="",
            resolution="",
            requirement=requirement,
            query=queryC,
            module=module,
        )
        for requirement in queryC.requirements
    ]
    for idx, result in enumerate(results[2]):
        assert result == query_c_expected_results[idx]

    # Validate the notifications
    if single_threaded:
        assert [call_args.args for call_args in status_func.call_args_list] == [
            (0, 0, 0, 0, 0),
            (1, 1, 0, 0, 0),
            (2, 1, 1, 0, 0),
            (3, 2, 1, 0, 0),
            (4, 3, 1, 0, 0),
            (4, 3, 1, 0, 0),
            (5, 4, 1, 0, 0),
            (6, 5, 1, 0, 0),
            (7, 6, 1, 0, 0),
            (8, 7, 1, 0, 0),
            (11, 7, 1, 0, 3),
        ]


# ----------------------------------------------------------------------
def test_ModuleNoData():
    module = MyModule(
        "MyModule",
        "The only module",
        ExecutionStyle.Sequential,
        [queryA, queryB, queryC],
        produce_data=False,
    )

    initial_data = module.GenerateInitialData({})
    assert initial_data is None


# ----------------------------------------------------------------------
def test_Module_SingleParallel():
    """Test parallel execution of the module when only 1 query is present."""
    # Since we have only one query with parallel execution style,
    # it should default to sequential processing.
    module = MyModule(
        "MyModule",
        "The only module",
        ExecutionStyle.Sequential,
        [queryA],
        produce_data=True,
    )

    status_func = Mock()

    initial_data = module.GenerateInitialData({})
    results = module.Evaluate(
        initial_data,
        {},
        status_func,
        max_num_threads=1,
    )
    assert results[0][0].result == EvaluateResult.Success
    assert results[0][0].context is None
    assert results[0][0].resolution is None
    assert results[0][0].rationale is None
    assert results[0][0].requirement is requirementA1
    assert results[0][0].query is queryA
    assert results[0][0].module is module


# ----------------------------------------------------------------------
def test_ProvidedDoneManager():
    """Test for when a DoneManager is provided to the ParallelSequentialProcessor."""
    module = MyModule(
        "MyModule",
        "The only module",
        ExecutionStyle.Sequential,
        [queryA],
        produce_data=True,
    )
    dm_and_sink = GenerateDoneManagerAndContent()
    dm = cast(DoneManager, next(dm_and_sink))
    result = ParallelSequentialProcessor(module.queries, lambda query: (0, []), dm)
    assert isinstance(result, list)


# ----------------------------------------------------------------------
class TestProcessRequriements:
    # ----------------------------------------------------------------------
    def test_NoIncludes(self):
        module = MyModule(
            "MyModule",
            "",
            ExecutionStyle.Sequential,
            [copy.deepcopy(queryA)],
            produce_data=False,
        )

        assert module.GetNumRequirements() == 4
        assert module.queries

        module.ProcessRequirements(
            set(),
            set(),
        )

        # The requires_explicit_include requirements should be removed
        assert module.GetNumRequirements() == 3
        assert module.queries

        module.ProcessRequirements(set(), {"RequirementA2"})
        assert module.GetNumRequirements() == 2
        assert module.queries

        module.ProcessRequirements(set(), {"RequirementA3", "RequirementA4"})
        assert module.GetNumRequirements() == 0
        assert not module.queries

    # ----------------------------------------------------------------------
    def test_WithIncludes(self):
        module = MyModule(
            "MyModule",
            "",
            ExecutionStyle.Sequential,
            [copy.deepcopy(queryA)],
            produce_data=False,
        )

        assert module.GetNumRequirements() == 4
        assert module.queries

        module.ProcessRequirements(
            {"RequirementA1"},
            set(),
        )

        assert module.GetNumRequirements() == 4
        assert module.queries

        module.ProcessRequirements(
            {"RequirementA1"},
            {"RequirementA3", "RequirementA4"},
        )

        assert module.GetNumRequirements() == 2
        assert module.queries



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the Description object."""

import textwrap
from typing import Any

import typer
from dbrownell_Common.TyperEx import TypeDefinitionItemType  # type: ignore[import-untyped]
from dbrownell_Common.Types import override  # type: ignore[import-untyped]

from RepoAuditor.Requirement import EvaluateResult, ExecutionStyle, Requirement


# ----------------------------------------------------------------------
class Description(Requirement):
    """Requirement to validate a repository's description."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "Description",
            "Validates a repository's description.",
            ExecutionStyle.Parallel,
            textwrap.dedent(
                """\
                1) Visit '{session.github_url}'
                2) Locate the 'About' section
                3) Click the gear icon
                4) Add a description
                """,
            ),
            textwrap.dedent(
                """\
                The default behavior is to require a repository description.

                Reasons for this Default
                ------------------------
                - A short description is helpful with browsing for repositories associated with a user or organization.

                Reasons to Override this Default
                --------------------------------
                - You do not want to spend time writing a short description.
                """,
            ),
        )

    # ----------------------------------------------------------------------
    @override
    def GetDynamicArgDefinitions(self) -> dict[str, TypeDefinitionItemType]:
        """Get the definitions for the arguments to this requirement."""
        return {
            "allow-empty": (
                bool,
                typer.Option(
                    default=False,
                    help="Allow an empty repository description.",
                ),
            ),
        }

    # ----------------------------------------------------------------------
    @override
    def _EvaluateImpl(
        self,
        query_data: dict[str, Any],
        requirement_args: dict[str, Any],
    ) -> Requirement.EvaluateImplResult:
        standard_data = query_data["standard"]

        # If result is None, it means the description is empty.
        # It is not a permissions issue, since "Read Metadata" is a mandatory permission for all PATs
        result = standard_data.get("description", None)
        expect_description = not requirement_args["allow-empty"]

        if expect_description and not result:
            return Requirement.EvaluateImplResult(
                EvaluateResult.Error,
                "The repository's description must be not be empty.",
                provide_resolution=True,
                provide_rationale=expect_description,
            )

        return Requirement.EvaluateImplResult(EvaluateResult.Success, None)



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the CodeOwners requirement."""

import textwrap

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import ExistsRequirementImpl


class CodeOwners(ExistsRequirementImpl):
    """Validates that CODEOWNERS file is configured."""

    def __init__(self) -> None:
        super().__init__(
            "CodeOwners",
            "CODEOWNERS",
            [
                ".github/CODEOWNERS",
                "docs/CODEOWNERS",
                "CODEOWNERS",
            ],
            textwrap.dedent(
                """\
                1) Create a CODEOWNERS file in one of these locations:
                   - .github/CODEOWNERS
                   - docs/CODEOWNERS
                   - CODEOWNERS

                Example CODEOWNERS file format:
                # Syntax example:
                # directory/ @username
                # *.js @username
                # /docs/ @username @username2
                """
            ),
            textwrap.dedent(
                """\
                CODEOWNERS files define who is responsible for code review.

                Benefits:
                - Automatic reviewer assignment
                - Clear code ownership
                - Streamlined review process
                """
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the IssueTemplates requirement."""

from RepoAuditor.Plugins.CommunityStandards.Impl.ExistsRequirementImpl import ExistsRequirementImpl


class IssueTemplates(ExistsRequirementImpl):
    """Requirement that checks for the presence of issue templates."""

    def __init__(self) -> None:
        super().__init__(
            "IssueTemplates",
            "ISSUE_TEMPLATES",
            [
                ".github/ISSUE_TEMPLATE.md",
                "docs/ISSUE_TEMPLATE.md",
                "ISSUE_TEMPLATE.md",
                ".github/issue_template.md",
                "docs/issue_template.md",
                "issue_template.md",
                # Directories
                ".github/ISSUE_TEMPLATE",
            ],
            "Add issue templates to your repository in the .github/ISSUE_TEMPLATE directory.",
            "Issue templates help contributors create high-quality issues by providing a structured format.",
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Contains the DefaultBranch object."""

import textwrap

from RepoAuditor.Plugins.GitHub.StandardRequirements.Impl.StandardValueRequirementImpl import (
    StandardValueRequirementImpl,
)


# ----------------------------------------------------------------------
class DefaultBranch(StandardValueRequirementImpl):
    """Requirement of a default branch in the repository."""

    # ----------------------------------------------------------------------
    def __init__(self) -> None:
        super().__init__(
            "DefaultBranch",
            "main",
            "settings",
            "Default Branch",
            None,
            lambda data: data["standard"].get("default_branch", None),
            textwrap.dedent(
                """\
                The default behavior is not name the mainline/base/default branch "main".

                Reasons for this Default
                ------------------------
                - Eliminate divisive language in favor of non-divisive language, this includes eliminating
                  the use of terms that were inappropriately and offensively taken from slavery including
                  the elimination of the term master in favor of main.

                  https://www.linkedin.com/pulse/technology-notes-how-tos-infrastructure-git-master-main-eldritch/

                Reasons to Override this Default
                --------------------------------
                - You are validating a legacy repository that still uses 'master' as the mainline/base/default branch.
                """,
            ),
        )



# -------------------------------------------------------------------------------
# |
# |  Copyright (c) 2024 Scientific Software Engineering Center at Georgia Tech
# |  Distributed under the MIT License.
# |
# -------------------------------------------------------------------------------
"""Unit tests for Protected.py"""

import pytest

from RepoAuditor.Plugins.GitHub.DefaultBranchRequirements.Protected import Protected
from RepoAuditor.Requirement import EvaluateResult


@pytest.fixture
def query_data(session):
    return {
        "session": session,
        "default_branch_data": {
            "protected": True,
        },
    }


@pytest.fixture
def requirement():
    return Protected()


class TestProtected:
    def test_Incomplete(self, requirement, query_data):
        """Test for incomplete result"""
        query_data["default_branch_data"] = {}
        requirement_args = {}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Warning
        assert "Incomplete data was encountered" in result.context

    def test_UnprotectedMainlineBranch(self, requirement, query_data):
        """Test for unprotected mainline branch."""
        query_data["default_branch_data"] = {"protected": True}
        # Don't protect mainline branch
        requirement_args = {"disabled": True}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert result.rationale is None
        assert result.resolution is not None

    def test_ProtectedMainlineBranch(self, requirement, query_data):
        """Test for protected mainline branch"""
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Success
        assert result.rationale is None
        assert result.resolution is None

    def test_is_protected_False(self, requirement, query_data):
        """Test when is_protected is False"""
        query_data["default_branch_data"]["protected"] = False
        requirement_args = {"disabled": False}
        result = requirement.Evaluate(query_data, requirement_args)
        assert result.result == EvaluateResult.Error
        assert result.rationale is not None
        assert result.resolution is not None
